import pytest
import numpy as np

from psyneulink.core.components.mechanisms.processing.objectivemechanism import ObjectiveMechanism
from psyneulink.core.components.mechanisms.processing.processingmechanism import ProcessingMechanism
from psyneulink.core.components.mechanisms.processing.transfermechanism import TransferMechanism
from psyneulink.core.components.mechanisms.modulatory.control import OptimizationControlMechanism
from psyneulink.library.components.mechanisms.processing.objective.comparatormechanism import ComparatorMechanism
from psyneulink.core.components.functions.objectivefunctions import Distance
from psyneulink.core.components.functions.optimizationfunctions import GridSearch, MINIMIZE
from psyneulink.core.components.functions.transferfunctions import GaussianDistort, Exponential
from psyneulink.core.components.ports.modulatorysignals.controlsignal import ControlSignal
from psyneulink.core.components.ports.inputport import SHADOW_INPUTS
from psyneulink.core.compositions.composition import Composition, NodeRole
from psyneulink.core.globals.keywords import VARIANCE, NORMED_L0_SIMILARITY

@pytest.mark.model
@pytest.mark.benchmark(group="Greedy Agent")
@pytest.mark.parametrize("mode", ['Python',
    pytest.param('LLVM', marks=[pytest.mark.llvm]),
    pytest.param('LLVMExec', marks=[pytest.mark.llvm]),
    pytest.param('LLVMRun', marks=[pytest.mark.llvm]),
    pytest.param('PTXExec', marks=[pytest.mark.llvm, pytest.mark.cuda]),
    pytest.param('PTXRun', marks=[pytest.mark.llvm, pytest.mark.cuda])])
def test_simplified_greedy_agent(benchmark, mode):
    # These should probably be replaced by reference to ForagerEnv constants:
    obs_len = 2
    action_len = 2
    player_coord_idx = slice(0,2)
    predator_coord_idx = slice(3,5)
    prey_coord_idx = slice(6,8)
    player_value_idx = 2
    predator_value_idx = 5
    prey_value_idx = 8

    player_len = prey_len = predator_len = obs_len

# The original needs GaussianDistort
#    player = ProcessingMechanism(size=prey_len, function=GaussianDistort, name="PLAYER OBS")
#    prey = ProcessingMechanism(size=prey_len, function=GaussianDistort, name="PREY OBS")
    player = TransferMechanism(size=prey_len, name="PLAYER OBS")
    prey = TransferMechanism(size=prey_len, name="PREY OBS")

    # Use ComparatorMechanism to compute direction of action as difference of coordinates between player and prey:
    # note: unitization is done in main loop, to allow compilation of LinearCombination function) (TBI)
    greedy_action_mech = ComparatorMechanism(name='MOTOR OUTPUT',sample=player,target=prey)

    agent_comp = Composition(name='PREDATOR-PREY COMPOSITION')
    agent_comp.add_node(player)
    agent_comp.add_node(prey)
    agent_comp.add_node(greedy_action_mech)

    # Projections to greedy_action_mech were created by assignments of sample and target args in its constructor,
    #  so just add them to the Composition).
    for projection in greedy_action_mech.projections:
        agent_comp.add_projection(projection)

    run_results = agent_comp.run(inputs={player:[[619,177]],
                                         prey:[[419,69]]},
                                 bin_execute=mode)
    assert np.allclose(run_results, [[-200, -108]])
    benchmark(agent_comp.run, **{'inputs':{
        player:[[619,177]],
        prey:[[419,69]],
        }, 'bin_execute':mode})

@pytest.mark.model
@pytest.mark.benchmark(group="Greedy Agant Random")
@pytest.mark.parametrize("mode", ['Python',
    pytest.param('LLVM', marks=[pytest.mark.llvm]),
    pytest.param('LLVMExec', marks=[pytest.mark.llvm]),
    pytest.param('LLVMRun', marks=[pytest.mark.llvm]),
    pytest.param('PTXExec', marks=[pytest.mark.llvm, pytest.mark.cuda]),
    pytest.param('PTXRun', marks=[pytest.mark.llvm, pytest.mark.cuda])])
def test_simplified_greedy_agent_random(benchmark, mode):
    # These should probably be replaced by reference to ForagerEnv constants:
    obs_len = 2
    action_len = 2
    player_coord_idx = slice(0,2)
    predator_coord_idx = slice(3,5)
    prey_coord_idx = slice(6,8)
    player_value_idx = 2
    predator_value_idx = 5
    prey_value_idx = 8

    player_len = prey_len = predator_len = obs_len

    player = ProcessingMechanism(size=prey_len, function=GaussianDistort, name="PLAYER OBS")
    prey = ProcessingMechanism(size=prey_len, function=GaussianDistort, name="PREY OBS")

    # Use ComparatorMechanism to compute direction of action as difference of coordinates between player and prey:
    # note: unitization is done in main loop, to allow compilation of LinearCombination function) (TBI)
    greedy_action_mech = ComparatorMechanism(name='MOTOR OUTPUT',sample=player,target=prey)

    agent_comp = Composition(name='PREDATOR-PREY COMPOSITION')
    agent_comp.add_node(player)
    agent_comp.add_node(prey)
    agent_comp.add_node(greedy_action_mech)

    # Projections to greedy_action_mech were created by assignments of sample and target args in its constructor,
    #  so just add them to the Composition).
    for projection in greedy_action_mech.projections:
        agent_comp.add_projection(projection)

    run_results = agent_comp.run(inputs={player:[[619,177]],
                                         prey:[[419,69]]},
                                 bin_execute=mode)
    # KDM 12/4/19: modified results due to global seed offset of
    # GaussianDistort assignment.
    # to produce old numbers, run get_global_seed once before creating
    # each Mechanism with GaussianDistort above
    assert np.allclose(run_results, [[-199.5484223217141, -107.79361870517444]])
    benchmark(agent_comp.run, **{'inputs':{
        player:[[619,177]],
        prey:[[419,69]],
        }, 'bin_execute':mode})

@pytest.mark.model
@pytest.mark.benchmark(group="Predator Prey")
@pytest.mark.parametrize("mode", ['Python',
     pytest.param('Python-PTX', marks=[pytest.mark.llvm, pytest.mark.cuda]),
     pytest.param('LLVM', marks=[pytest.mark.llvm]),
     pytest.param('LLVMExec', marks=[pytest.mark.llvm]),
     pytest.param('LLVMRun', marks=[pytest.mark.llvm]),
     pytest.param('PTXExec', marks=[pytest.mark.llvm, pytest.mark.cuda]),
     pytest.param('PTXRun', marks=[pytest.mark.llvm, pytest.mark.cuda]),
])
@pytest.mark.parametrize("samples", [[0,10],
    pytest.param([a / 10.0 for a in range(0, 101)]),
    pytest.param([0,3,6,10], marks=pytest.mark.stress),
    pytest.param([0,2,4,6,8,10], marks=pytest.mark.stress),
], ids=lambda x: len(x))
def test_predator_prey(benchmark, mode, samples):
    if len(samples) > 10 and mode not in {"LLVMRun", "Python-PTX"}:
        pytest.skip("This test takes too long")
    # OCM default mode is Python
    mode, ocm_mode = (mode + "-Python").split('-')[0:2]
    benchmark.group = "Predator-Prey " + str(len(samples))
    obs_len = 3
    obs_coords = 2
    player_idx = 0
    player_obs_start_idx = player_idx * obs_len
    player_value_idx = player_idx * obs_len + obs_coords
    player_coord_slice = slice(player_obs_start_idx,player_value_idx)
    predator_idx = 1
    predator_obs_start_idx = predator_idx * obs_len
    predator_value_idx = predator_idx * obs_len + obs_coords
    predator_coord_slice = slice(predator_obs_start_idx,predator_value_idx)
    prey_idx = 2
    prey_obs_start_idx = prey_idx * obs_len
    prey_value_idx = prey_idx * obs_len + obs_coords
    prey_coord_slice = slice(prey_obs_start_idx,prey_value_idx)

    player_len = prey_len = predator_len = obs_coords

    # Perceptual Mechanisms
    player_obs = ProcessingMechanism(size=prey_len, function=GaussianDistort, name="PLAYER OBS")
    prey_obs = ProcessingMechanism(size=prey_len, function=GaussianDistort, name="PREY OBS")
    predator_obs = TransferMechanism(size=predator_len, function=GaussianDistort, name="PREDATOR OBS")

    # Action Mechanism
    #    Use ComparatorMechanism to compute direction of action as difference of coordinates between player and prey:
    #    note: unitization is done in main loop, to allow compilation of LinearCombination function) (TBI)
    greedy_action_mech = ComparatorMechanism(name='ACTION',sample=player_obs,target=prey_obs)

    # Create Composition
    agent_comp = Composition(name='PREDATOR-PREY COMPOSITION')
    agent_comp.add_node(player_obs)
    agent_comp.add_node(predator_obs)
    agent_comp.add_node(prey_obs)
    agent_comp.add_node(greedy_action_mech)
    agent_comp.exclude_node_roles(predator_obs, NodeRole.OUTPUT)

    ocm = OptimizationControlMechanism(features={SHADOW_INPUTS: [player_obs, predator_obs, prey_obs]},
                                       agent_rep=agent_comp,
                                       function=GridSearch(direction=MINIMIZE,
                                                           save_values=True),

                                       objective_mechanism=ObjectiveMechanism(function=Distance(metric=NORMED_L0_SIMILARITY),
                                                                              monitor=[
                                                                                  player_obs,
                                                                                  prey_obs
                                                                              ]),
                                       control_signals=[ControlSignal(modulates=(VARIANCE,player_obs),
                                                                      allocation_samples=samples),
                                                        ControlSignal(modulates=(VARIANCE,predator_obs),
                                                                      allocation_samples=samples),
                                                        ControlSignal(modulates=(VARIANCE,prey_obs),
                                                                      allocation_samples=samples)
                                                        ],
                                       )
    agent_comp.add_controller(ocm)
    agent_comp.enable_controller = True
    ocm.comp_execution_mode = ocm_mode

    input_dict = {player_obs:[[1.1576537,  0.60782117]],
                  predator_obs:[[-0.03479106, -0.47666293]],
                  prey_obs:[[-0.60836214,  0.1760381 ]],
                 }
    run_results = agent_comp.run(inputs=input_dict, num_trials=2, bin_execute=mode)

    if len(samples) == 2:
        # KDM 12/4/19: modified results due to global seed offset of
        # GaussianDistort assignment.
        # to produce old numbers, run get_global_seed once before creating
        # each Mechanism with GaussianDistort above
        assert np.allclose(run_results[0], [[-10.06333025,   2.4845505 ]])
        if mode == 'Python':
            assert np.allclose(ocm.feature_values, [[ 1.1576537,   0.60782117],
                                                    [-0.03479106, -0.47666293],
                                                    [-0.60836214,  0.1760381 ]])

    benchmark(agent_comp.run, inputs=input_dict, bin_execute=mode)



#
# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
#
# *******************************************   SAMPLER CLASSES ********************************************************
"""

* `SampleSpec`
* `SampleIterator`

"""

import numpy as np

import typecheck as tc
from collections.abc import Iterator
from inspect import isclass
from decimal import Decimal, getcontext
from numbers import Number


__all__ = ['SampleSpec', 'SampleIterator']


# Number of decimal places used to set precision of decimal module
SAMPLE_SPEC_PRECISION = 16


def _validate_function(source, function):
    """Ensure function specification is appropriate for SampleIterator"""
    source_name = source.__class__.__name__
    try:
        result = function()
    except:
        raise SampleIteratorError("Specification of function for {} ({}) could not be called as a function".
                                  format(source_name, function.__class__.__name__))
    if result is None:
        raise SampleIteratorError("Function specified for {} ({}) does not return a result)".
                                  format(source_name, repr(function)))
    if not isinstance(result, Number):
        raise SampleIteratorError("Function specified for {} ({}) does not return a number)".
                                  format(source_name, repr(function)))


class SampleIteratorError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value


class SampleSpec():
    """
    SampleSpec(      \
    start=None,      \
    stop=None,       \
    step=None,       \
    num=None,        \
    function=None    \
    precision=None   \
    custom_spec=None \
    )

    Specify the information needed to create a SampleIterator that will either (1) generate discrete values in a range;
    (2) call a function can generate continous values within the range; (3) pass a custom spec to SampleIterator

    (1) Generate values in a range by explicitly specifying a finite regular sequence of values, using an appropriate
        combination of the **start**, **stop**, **step** and/or **num**arguments.

        * if **start**, **stop**, and **step** are specified, the behavior is similar to `Numpy's arange
          <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.arange.html>`_. Calling
          `next <SampleIterator.__next__>`first returns **start**. Each subsequent call to next returns
          **start** :math:`+` **step** :math:`*` current_step.
          Iteration stops when the current value exceeds the **stop** value.  If any of the specified arguments are
          floats, precision determines the number of decimal places used for rounding to ensure num is a int.

        * if **start**, **stop**, and **num** are specified, the behavior is similar to `Numpy's linspace
          <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.linspace.html>`_. Calling
          `next <SampleIterator.__next__>` first returns **start**. Each subsequent call to next returns
          **start** :math:`+` step :math:`*` current_step, where step is set to :math:`\\frac{stop-start)}{num - 1}`.
          Iteration stops when **num** is reached, or the current value exceeds the **stop** value.

        * if **start**, **stop*, **step**, and **num** are all specified, then **step** and **num** must be compatible.

    (2) Specify a function, that is called repeatedly to generate a sequence of values.

        * if **num** is specified, the **function** is called once on each iteration until **num** iterations are
          complete.

        * if **num** is not specified, the **function** is called once on each iteration, and iteration may continue
          indefintely.

    (3) Specify a custom_spec, that is passed to SampleIterator unmodified.

        * this is useful if the receiver of the SampleIterator uses a custom format or external functions to generate
          samples;  the format must, of course, be one expected by the receiver.

    .. note::
        * Some OptimizationFunctions may require that their SampleIterators have a "num" attribute.
        * The Python decimal module is used to implement **precision** for rounding.


    Arguments
    ---------

    start : int or float
        first sample in the sequence.

    stop : int or float
        maximum value of last sample in the sequence.

    step : int or float
        space between samples in sequence.

    num :  int
        number of samples.

    function : function or function class
        function to be called on each iteration. Must return one sample each time it is called.

    precision : int default 16
        number of decimal places used for rounding in floating point operations.

    custom_spec : anything

    Attributes
    ----------

    start : float
        first sample in the sequence.

    stop : float
        maximum value of last sample in the sequence.

    step : float
        space between samples in sequence.

    num :  int
        number of samples.

    function : function
        Function to be called on each iteration. Must return one sample.

    custom_spec : anything
        specification in a format recognized by receiver of SampleIterator.

    """

    @tc.typecheck
    def __init__(self,
                 start:tc.optional(tc.any(int, float))=None,
                 stop:tc.optional(tc.any(int, float))=None,
                 step:tc.optional(tc.any(int, float))=None,
                 num:tc.optional(int)=None,
                 function:tc.optional(callable)=None,
                 precision:tc.optional(int)=None,
                 custom_spec = None
                 ):

        self.custom_spec = custom_spec

        if self.custom_spec:
            # Assumes receiver of SampleIterator will get this and know what to do with it,
            #   therefore no other attributes are needed and, to avoid confusion, they should not be available;
            #   so just assign and return.
            return

        self._precision = precision or SAMPLE_SPEC_PRECISION
        # Save global precision for later restoration
        _global_precision = getcontext().prec
        # Set SampleSpec precision
        getcontext().prec = self._precision

        if function is None:
            if start is None or stop is None:
                raise SampleIteratorError("If 'function' is not specified, then 'start' and 'stop' must be "
                                                "specified.")
            if num is None and step is not None:
                num = int(Decimal(1.0) + Decimal(stop - start) / Decimal(step))
                # num = 1.0 + (stop - start) / step
            elif step is None and num is not None:
                step = (Decimal(stop) - Decimal(start)) / (num - 1)
            elif num is None and step is None:
                raise SampleIteratorError("Must specify one of {}, {} or {}."
                                                .format(repr('step'), repr('num'), repr('function')))
            else:
                if not np.isclose(num, 1.0 + (stop - start) / step):
                    raise SampleIteratorError("The {} ({}) and {} ({}} values specified are not comaptible."
                                                    .format(repr('step'), step, repr('num'), num))

        elif callable(function):
            _validate_function(self, function)

            if start is not None:
                raise SampleIteratorError("Only one of {} ({}) and {} ({}} may be specified."
                                                .format(repr('start'), start, repr('function'), function))
            if step is not None:
                raise SampleIteratorError("Only one of {} ({}) and {} ({}} may be specified."
                                                .format(repr('step'), step, repr('function'), function))
        else:
            raise SampleIteratorError("{} is not a valid function for {}."
                                            .format(function, self.__name__))

        # FIX: ELIMINATE WHEN UPGRADING TO PYTHON 3.5.2 OR 3.6, (AND USING ONE OF THE TYPE VERSIONS COMMENTED OUT ABOVE)
        # Validate entries of specification
        #
        self.start = start
        self.stop = stop
        self.step = step
        self.num = num
        self.function = function

        # Restore global precision
        getcontext().prec = _global_precision


allowable_specs = (tuple, list, np.array, range, np.arange, callable, SampleSpec)
def is_sample_spec(spec):
    if spec is None or type(spec) in allowable_specs:
        return True
    return False


class SampleIterator(Iterator):
    """
    SampleIterator(               \
    specification                 \
    )

    Create an iterator that returns the next sample from a sequence on each call to `next <SampleIterator.__next__>`.
    (e.g., when next(<SampleIterator>) is called)

    The pattern of the sequence depends on the **specification**, which may be a list, nparray, range,
    function, or a SampleSpec. Most of the patterns depend on the "current_step," which is incremented on each
    iteration, and set to zero when the iterator is reset.

    +--------------------------------+-------------------------------------------+------------------------------------+
    | **Specification**              |  **what happens on each iteration**       | **StopIteration condition**        |
    +--------------------------------+-------------------------------------------+------------------------------------+
    | list, nparray                  | look up the item with index current_step  | list/array                         |
    +--------------------------------+-------------------------------------------+------------------------------------+
    | range, np.arange               | start + step*current_step                 | range stop value is reached        |
    +--------------------------------+-------------------------------------------+------------------------------------+
    | callable                       | call callable                             | iteration does not stop            |
    +--------------------------------+-------------------------------------------+------------------------------------+
    | SampleSpec(start, stop, step)  | start + step*current_step                 | current_step = num or value > stop |
    +--------------------------------+-------------------------------------------+------------------------------------+
    | SampleSpec(start, stop, num)   | start + step*current_step                 | current_step = num or value > stop |
    +--------------------------------+-------------------------------------------+------------------------------------+
    | SampleSpec(function, num)      | call function                             | current_step = num                 |
    +--------------------------------+-------------------------------------------+------------------------------------+
    | SampleSpec(function)           | call function                             | iteration does not stop            |
    +--------------------------------+-------------------------------------------+------------------------------------+

    .. note::
        We recommend reserving the list/nparray option for cases in which the samples do not have a pattern that can be
        represented by a SampleSpec, or the number of samples is small. The list/nparray option requires all of the
        samples to be stored and looked up, while the SampleSpec options generate samples as needed.

    """

    @tc.typecheck
    def __init__(self,
                 specification:tc.any(*allowable_specs)):
        """

        Arguments
        ---------

        specification : list, array, range, callable or SampleSpec
            specifies what to use for `generate_current_value <SampleIterator.generate_current_value>`.

        Attributes
        ----------

        start : scalar or None
            first sample in the sequence, or None

        stop : scalar or None
            last sample in the sequence, or None

        step : scalar
            space between samples in the sequence, or None

        num : int or None
            number of values returned before the iterator stops. If None, the iterator may be called indefinitely.

        current_step : int
            number of current iteration

        specification :  list, np.array, range, np.arange, callable, SampleSpec
            original argument provided in constructor;
            useful for passing application-specific forms of specification directly to receiver of SampleIterator.

        Returns
        -------

        List(self) : list
        """

        # FIX: DEAL WITH head?? OR SIMPLY USE CURRENT_STEP?
        # FIX Are nparrays allowed? Below assumes one list dimension. How to handle nested arrays/lists?
        self.specification = specification

        if isinstance(specification, (tuple, range)):
            specification = list(specification)

        elif callable(specification):
            if isclass(specification):
                specification = specification()
            _validate_function(self, specification)

            specification = SampleSpec(function=specification)

        if isinstance(specification, list):
            self.start = specification[0]
            self.stop = None
            self.step = None
            self.num = len(specification)
            self.generator = specification                       # the list

            def generate_current_value():                        # index into the list
                # KDM 12/11/19: for currently unknown and unreplicable
                # reasons, the checks in __next__ will fail to ensure
                # that self.current_step is less than the length of
                # self.generator, and an IndexError will be thrown
                # occurs in tests/log/test_log.py::TestFiltering
                # and tests/models/test_greedy_agent.py due to
                # GridSearch use
                try:
                    return self.generator[self.current_step]
                except IndexError:
                    raise StopIteration

        elif isinstance(specification, SampleSpec):

            if specification.custom_spec:
                # Assumes receiver of SampleIterator will get this and know what to do with it,
                #   therefore no other attributes are needed and, to avoid confusion, they should not be available;
                #   so just return.
                return

            if specification.function is None:
                self.start = specification.start
                self.stop = specification.stop
                # self.step = Fraction(specification.step)
                self.step = specification.step
                self.num = specification.num
                self.generator = None                    # ??

                def generate_current_value():   # return next value in range
                    # Save global precision for later restoration
                    _global_precision = getcontext().prec
                    # Set SampleSpec precision
                    getcontext().prec = specification._precision
                    return_value = float(Decimal(self.start) + Decimal(self.step) * Decimal(self.current_step))
                    # Restore global precision
                    getcontext().prec = _global_precision
                    return return_value

            elif callable(specification.function):
                self.start = 0
                self.stop = None
                self.step = 1
                self.current_step = 0
                self.num = specification.num
                self.head = self.start
                self.generator = specification.function

                def generate_current_value():  # call function
                    return self.generator()

            else:
                assert False, 'PROGRAM ERROR: {} item of {} passed to specification arg of {} ' \
                              'is not an iterator or a function_type'.\
                              format(repr('function'), SampleSpec.__name__, self.__class__.__name__)

        else:
            assert False, 'PROGRAM ERROR: {} argument of {} must be one of the following: {}'.\
                          format(repr('specification'), self.__class__.__name__,
                                 (', ').join([i.__name__ for i in allowable_specs]))

        self.current_step = 0
        self.head = self.start
        self.generate_current_value = generate_current_value

    def __next__(self):
        """

        :return:
        Sample value for the current iteration.
        """
        if self.num is None:
            current_value = self.generate_current_value()
            if hasattr(self, 'stop'):
                if self.stop is not None:
                    if current_value <= self.stop:
                        return current_value
                    else:
                        raise StopIteration
            return current_value
        if self.current_step < self.num:
            current_value = self.generate_current_value()
            self.current_step += 1
            if hasattr(self, 'stop'):
                if self.stop is not None:
                    if current_value <= self.stop:
                        return current_value
                    else:
                        raise StopIteration
            return current_value

        else:
            raise StopIteration

    def __iter__(self):
        self.current_step = 0
        return self

    def __call__(self):
        return list(self)

    def reset(self, head=None):
        """Reset iterator to a specified item
        If called with None, resets to first item (if `generator <SampleIterators.generator>` is a list or
        deterministic function.
        """

        self.current_step = 0
        self.head = head or self.start



import numpy as np
import pytest

import psyneulink as pnl
import psyneulink.core.llvm as pnlvm

from psyneulink.core.compositions.composition import Composition
from psyneulink.core.components.functions.function import FunctionError
from psyneulink.core.components.functions.distributionfunctions import NormalDist
from psyneulink.core.components.functions.statefulfunctions.integratorfunctions import \
    SimpleIntegrator, AdaptiveIntegrator, DriftDiffusionIntegrator, OrnsteinUhlenbeckIntegrator, \
    FitzHughNagumoIntegrator, AccumulatorIntegrator, LeakyCompetingIntegrator, DualAdaptiveIntegrator
from psyneulink.core.components.functions.transferfunctions import Linear
from psyneulink.core.components.mechanisms.mechanism import MechanismError
from psyneulink.core.components.mechanisms.processing.integratormechanism import \
    IntegratorMechanism, IntegratorMechanismError
from psyneulink.core.globals.context import Context
from psyneulink.core.scheduling.condition import AtTrial
from psyneulink.core.scheduling.condition import Never


class TestReset:
    def test_FitzHughNagumo_valid(self):
        I = IntegratorMechanism(name="I",
                                function=FitzHughNagumoIntegrator())
        I.reset_stateful_function_when = Never()
        I.execute(1.0)

        assert np.allclose([[0.05127053]], I.value[0])
        assert np.allclose([[0.00279552]], I.value[1])
        assert np.allclose([[0.05]], I.value[2])

        I.function.reset(0.01, 0.02, 0.03)

        assert np.allclose(0.01, I.function.value[0])
        assert np.allclose(0.02, I.function.value[1])
        assert np.allclose(0.03, I.function.value[2])

        assert np.allclose([[0.05127053]], I.value[0])
        assert np.allclose([[0.00279552]], I.value[1])
        assert np.allclose([[0.05]], I.value[2])

        assert np.allclose([[0.05127053]], I.output_ports[0].value)

        I.execute(1.0)

        assert np.allclose([[0.06075727]], I.value[0])
        assert np.allclose([[0.02277156]], I.value[1])
        assert np.allclose([[0.08]], I.value[2])

        assert np.allclose([[0.06075727]], I.output_ports[0].value)

        # I.reset(new_previous_v=0.01, new_previous_w=0.02, new_previous_time=0.03)
        I.reset(0.01, 0.02, 0.03)

        assert np.allclose(0.01, I.value[0])
        assert np.allclose(0.02, I.value[1])
        assert np.allclose(0.03, I.value[2])

        assert np.allclose(0.01, I.output_ports[0].value)
        # assert np.allclose(0.01, I.output_port.value[0])
        # assert np.allclose(0.02, I.output_port.value[1])
        # assert np.allclose(0.03, I.output_port.value[2])

    def test_AGTUtility_valid(self):
        I = IntegratorMechanism(name="I",
                                function=DualAdaptiveIntegrator())
        I.reset_stateful_function_when = Never()
        assert np.allclose([[0.0]], I.function.previous_short_term_avg)
        assert np.allclose([[0.0]], I.function.previous_long_term_avg)

        I.function.reset(0.2, 0.8)

        assert np.allclose([[0.2]], I.function.previous_short_term_avg)
        assert np.allclose([[0.8]], I.function.previous_long_term_avg)

        I.function.reset()

        assert np.allclose([[0.0]], I.function.previous_short_term_avg)
        assert np.allclose([[0.0]], I.function.previous_long_term_avg)

        I.reset(0.3, 0.7)

        assert np.allclose([[0.3]], I.function.previous_short_term_avg)
        assert np.allclose([[0.7]], I.function.previous_long_term_avg)
        context = Context()
        print(I.value)
        print(I.function._combine_terms(0.3, 0.7, context))
        assert np.allclose(I.function._combine_terms(0.3, 0.7, context), I.value)

        I.reset()

        assert np.allclose([[0.0]], I.function.previous_short_term_avg)
        assert np.allclose([[0.0]], I.function.previous_long_term_avg)
        assert np.allclose(I.function._combine_terms(0.0, 0.0, context), I.value)

    def test_Simple_valid(self):
        I = IntegratorMechanism(
            name='IntegratorMechanism',
            function=SimpleIntegrator(
            ),
        )
        I.reset_stateful_function_when = Never()

        #  returns previous_value + rate*variable + noise
        # so in this case, returns 10.0
        I.execute(10)
        assert np.allclose(I.value, 10.0)
        assert np.allclose(I.output_port.value, 10.0)

        # reset function
        I.function.reset(5.0)
        assert np.allclose(I.function.value, 5.0)
        assert np.allclose(I.value, 10.0)
        assert np.allclose(I.output_ports[0].value, 10.0)

        # reset function without value spec
        I.function.reset()
        assert np.allclose(I.function.value, 0.0)
        assert np.allclose(I.value, 10.0)
        assert np.allclose(I.output_ports[0].value, 10.0)

        # reset mechanism
        I.reset(4.0)
        assert np.allclose(I.function.value, 4.0)
        assert np.allclose(I.value, 4.0)
        assert np.allclose(I.output_ports[0].value, 4.0)

        I.execute(1)
        assert np.allclose(I.value, 5.0)
        assert np.allclose(I.output_ports[0].value, 5.0)

        # reset mechanism without value spec
        I.reset()
        assert np.allclose(I.function.value, 0.0)
        assert np.allclose(I.value, 0.0)
        assert np.allclose(I.output_ports[0].value, 0.0)

    def test_Adaptive_valid(self):
        I = IntegratorMechanism(
            name='IntegratorMechanism',
            function=AdaptiveIntegrator(
                rate=0.5
            ),
        )

        #  returns (1-rate)*previous_value + rate*variable + noise
        # so in this case, returns 0.5*0 + 0.5*10 + 0 = 5.0
        I.execute(10)
        assert np.allclose(I.value, 5.0)
        assert np.allclose(I.output_port.value, 5.0)

        # reset function
        I.function.reset(1.0)
        assert np.allclose(I.function.value, 1.0)
        assert np.allclose(I.value, 5.0)
        assert np.allclose(I.output_ports[0].value, 5.0)

        # reset function without value spec
        I.function.reset()
        assert np.allclose(I.function.value, 0.0)
        assert np.allclose(I.value, 5.0)
        assert np.allclose(I.output_ports[0].value, 5.0)

        # reset mechanism
        I.reset(2.0)
        assert np.allclose(I.function.value, 2.0)
        assert np.allclose(I.value, 2.0)
        assert np.allclose(I.output_ports[0].value, 2.0)

        I.execute(1.0)
        #  (1-0.5)*2.0 + 0.5*1.0 + 0 = 1.5
        assert np.allclose(I.value, 1.5)
        assert np.allclose(I.output_ports[0].value, 1.5)

        # reset mechanism without value spec
        I.reset()
        assert np.allclose(I.function.value, 0.0)
        assert np.allclose(I.value, 0.0)
        assert np.allclose(I.output_ports[0].value, 0.0)

    def test_Accumulator_warning(self):
        regexp = r"AccumulatorIntegrator does not use its variable;  value passed .* will be ignored"
        with pytest.warns(UserWarning, match=regexp):
            I = AccumulatorIntegrator()
            I(1000)

    def test_Accumulator_valid(self):
        I = IntegratorMechanism(
            name='IntegratorMechanism',
            function=AccumulatorIntegrator(increment=1.0),
        )

        #  returns previous_value + rate + noise
        # so in this case, returns 0.0 + 1.0
        I.execute(1000)
        assert np.allclose(I.value, 1.0)
        assert np.allclose(I.output_port.value, 1.0)

        # reset function
        I.function.reset(2.0)
        assert np.allclose(I.function.value, 2.0)
        assert np.allclose(I.value, 1.0)
        assert np.allclose(I.output_ports[0].value, 1.0)

        # reset function without value spec
        I.function.reset()
        assert np.allclose(I.function.value, 0.0)
        assert np.allclose(I.value, 1.0)
        assert np.allclose(I.output_ports[0].value, 1.0)

        # reset mechanism
        I.reset(2.0)
        assert np.allclose(I.function.value, 2.0)
        assert np.allclose(I.value, 2.0)
        assert np.allclose(I.output_ports[0].value, 2.0)

        I.execute(1000)
        #  2.0 + 1.0 = 3.0
        assert np.allclose(I.value, 3.0)
        assert np.allclose(I.output_ports[0].value, 3.0)

        # reset mechanism without value spec
        I.reset()
        assert np.allclose(I.function.value, 0.0)
        assert np.allclose(I.value, 0.0)
        assert np.allclose(I.output_ports[0].value, 0.0)

    def test_OU_valid(self):
        I = IntegratorMechanism(
            name='IntegratorMechanism',
            function=OrnsteinUhlenbeckIntegrator(),
        )

        # previous_value + (decay * previous_value - rate * variable) * time_step_size + noise
        # decay=1.0, initializer=0.0, rate=1.0, time_step_size=1.0, noise=0.0
        # returns 0.0 + (1.0*0.0 - 1.0*10.0*1.0) + 0.0 = -10.0
        I.execute(2.0)
        assert np.allclose(I.value[0], -2.0)
        assert np.allclose(I.output_port.value, -2.0)

        # reset function
        I.function.reset(5.0, 0.0)
        assert np.allclose(I.function.value[0], 5.0)
        assert np.allclose(I.value[0], -2.0)
        assert np.allclose(I.output_ports[0].value, -2.0)

        # reset function without value spec
        I.function.reset()
        assert np.allclose(I.function.value[0], 0.0)
        assert np.allclose(I.value[0], -2.0)
        assert np.allclose(I.output_ports[0].value, -2.0)

        # reset mechanism
        I.reset(4.0, 0.0)
        assert np.allclose(I.function.value[0], 4.0)
        assert np.allclose(I.value[0], 4.0)
        assert np.allclose(I.output_ports[0].value, 4.0)

        I.execute(1.0)
        # 4.0 + (1.0 * 4.0 - 1.0 * 1.0) * 1.0 = 4 + 3 = 7
        assert np.allclose(I.value[0], 7.0)
        assert np.allclose(I.output_ports[0].value, 7.0)

        # reset mechanism without value spec
        I.reset()
        assert np.allclose(I.function.value[0], 0.0)
        assert np.allclose(I.value[0], 0.0)
        assert np.allclose(I.output_ports[0].value, 0.0)

    def test_Accumulator_valid(self):
        I = IntegratorMechanism(
            name='IntegratorMechanism',
            function=AccumulatorIntegrator(increment=0.1),
        )

        #  returns previous_value * rate + noise + increment
        # initializer = 0.0, rate = 1.0, noise = 0.0, increment = 0.1
        # returns 0.0*1.0 + 0.0 + 0.1 = 0.1
        I.execute(10000)
        assert np.allclose(I.value, 0.1)
        assert np.allclose(I.output_port.value, 0.1)

        # reset function
        I.function.reset(2.0)
        assert np.allclose(I.function.value, 2.0)
        assert np.allclose(I.value, 0.1)
        assert np.allclose(I.output_ports[0].value, 0.1)

        # reset function without value spec
        I.function.reset()
        assert np.allclose(I.function.value, 0.0)
        assert np.allclose(I.value, 0.1)
        assert np.allclose(I.output_ports[0].value, 0.1)

        # reset mechanism
        I.reset(5.0)
        assert np.allclose(I.function.value, 5.0)
        assert np.allclose(I.value, 5.0)
        assert np.allclose(I.output_ports[0].value, 5.0)

        I.execute(10000)
        #  5.0 * 1.0 + 0.0 + 0.1
        assert np.allclose(I.value, 5.1)
        assert np.allclose(I.output_ports[0].value, 5.1)

        # reset mechanism without value spec
        I.reset()
        assert np.allclose(I.function.value, 0.0)
        assert np.allclose(I.value, 0.0)
        assert np.allclose(I.output_ports[0].value, 0.0)

    def test_LCIIntegrator_valid(self):
        I = IntegratorMechanism(
            name='IntegratorMechanism',
            function=LeakyCompetingIntegrator(leak=1),
        )

        # previous_value + (new_value - rate*previous_value)*time_step_size + noise
        # initializer=0.0, rate=1.0, time_step_size=0.1, noise=0.0
        # returns 0.0 + (1.0*0.0 + 2.0)*0.1 = 2.0
        I.execute(2.0)
        assert np.allclose(I.value, 0.2)
        assert np.allclose(I.output_port.value, 0.2)

        # reset function
        I.function.reset(5.0)
        assert np.allclose(I.function.value, 5.0)
        assert np.allclose(I.value, 0.2)
        assert np.allclose(I.output_ports[0].value, 0.2)

        # reset function without value spec
        I.function.reset()
        assert np.allclose(I.function.value, 0.0)
        assert np.allclose(I.value, 0.2)
        assert np.allclose(I.output_ports[0].value, 0.2)

        # reset mechanism
        I.reset(4.0)
        assert np.allclose(I.function.value, 4.0)
        assert np.allclose(I.value, 4.0)
        assert np.allclose(I.output_ports[0].value, 4.0)

        I.execute(1.0)
        # 4.0 + (1.0*4.0 + 1.0)*0.1 + 0.0
        assert np.allclose(I.value, 3.7)
        assert np.allclose(I.output_ports[0].value, 3.7)

        # reset mechanism without value spec
        I.reset()
        assert np.allclose(I.function.value, 0.0)
        assert np.allclose(I.value, 0.0)
        assert np.allclose(I.output_ports[0].value, 0.0)

    def test_reset_not_integrator(self):

        with pytest.raises(MechanismError) as err_txt:
            I_not_integrator = IntegratorMechanism(function=Linear)
            I_not_integrator.execute(1.0)
            I_not_integrator.reset(0.0)
        assert "not allowed because this Mechanism is not stateful;" in str(err_txt.value)
        assert "it does not have an accumulator to reset" in str(err_txt.value)


VECTOR_SIZE=4

class TestIntegratorFunctions:

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_simple_integrator(self):
        I = IntegratorMechanism(
            function=SimpleIntegrator(
                initializer=10.0,
                rate=5.0,
                offset=10,
            )
        )
        # P = Process(pathway=[I])
        val = I.execute(1)
        assert val == 25

    @pytest.mark.mimo
    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    @pytest.mark.benchmark(group="IntegratorMechanism")
    @pytest.mark.parametrize('mode', ['Python',
                                      pytest.param('LLVM', marks=pytest.mark.llvm),
                                      pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda])])
    def test_integrator_multiple_input(self, benchmark, mode):
        I = IntegratorMechanism(
            function=Linear(slope=2.0, intercept=1.0),
            default_variable=[[1], [2]],
            input_ports=['a', 'b'],
        )
        if mode == 'Python':
            val = I.execute([[1], [2]])
            val = [x.value for x in I.output_ports]
            benchmark(I.execute, [[1], [2]])
        elif mode == 'LLVM':
            e = pnlvm.execution.MechExecution(I)
            val = e.execute([[1], [2]])
            benchmark(e.execute, [[1], [2]])
        elif mode == 'PTX':
            e = pnlvm.execution.MechExecution(I)
            val = e.cuda_execute([[1], [2]])
            benchmark(e.cuda_execute, [[1], [2]])

        assert np.allclose(val, [[3]])

    @pytest.mark.mimo
    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    @pytest.mark.benchmark(group="IntegratorMechanism")
    @pytest.mark.parametrize('mode', ['Python',
                                      pytest.param('LLVM', marks=pytest.mark.llvm),
                                      pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda])])
    def test_integrator_multiple_output(self, benchmark, mode):
        I = IntegratorMechanism(
            default_variable=[5],
            output_ports=[{pnl.VARIABLE: (pnl.OWNER_VALUE, 0)}, 'c'],
        )
        if mode == 'Python':
            val = I.execute([5])
            val = [x.value for x in I.output_ports]
            benchmark(I.execute, [5])
        elif mode == 'LLVM':
            e = pnlvm.execution.MechExecution(I)
            val = e.execute([5])
            benchmark(e.execute, [5])
        elif mode == 'PTX':
            e = pnlvm.execution.MechExecution(I)
            val = e.cuda_execute([5])
            benchmark(e.cuda_execute, [5])

        assert np.allclose(val, [[2.5], [2.5]])

    @pytest.mark.mimo
    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    @pytest.mark.benchmark(group="IntegratorMechanism")
    @pytest.mark.parametrize('mode', ['Python',
                                      pytest.param('LLVM', marks=pytest.mark.llvm),
                                      pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda])])
    def test_integrator_multiple_input_output(self, benchmark, mode):
        I = IntegratorMechanism(
            function=Linear(slope=2.0, intercept=1.0),
            default_variable=[[1], [2]],
            input_ports=['a', 'b'],
            output_ports=[{pnl.VARIABLE: (pnl.OWNER_VALUE, 1)},
                           {pnl.VARIABLE: (pnl.OWNER_VALUE, 0)}],
        )
        if mode == 'Python':
            val = I.execute([[1], [2]])
            val = [x.value for x in I.output_ports]
            benchmark(I.execute, [[1], [2]])
        elif mode == 'LLVM':
            e = pnlvm.execution.MechExecution(I)
            val = e.execute([[1], [2]])
            benchmark(e.execute, [[1], [2]])
        elif mode == 'PTX':
            e = pnlvm.execution.MechExecution(I)
            val = e.cuda_execute([[1], [2]])
            benchmark(e.cuda_execute, [[1], [2]])

        if mode == 'Python':
            val = [x.value for x in I.output_ports]
        assert np.allclose(val, [[5], [3]])

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    @pytest.mark.benchmark(group="IntegratorMechanism")
    @pytest.mark.parametrize('mode', ['Python',
                                      pytest.param('LLVM', marks=pytest.mark.llvm),
                                      pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda])])
    def test_FitzHughNagumo_simple_scalar(self, benchmark, mode):
        var = [1.0]
        I = IntegratorMechanism(name="I",
                                default_variable=[var],
                                function=FitzHughNagumoIntegrator())

        if mode == 'Python':
            val = I.execute(var)
            val = [x.value for x in I.output_ports]
            benchmark(I.execute, var)
        elif mode == 'LLVM':
            e = pnlvm.execution.MechExecution(I)
            val = e.execute(var)
            benchmark(e.execute, var)
        elif mode == 'PTX':
            e = pnlvm.execution.MechExecution(I)
            val = e.cuda_execute(var)
            benchmark(e.cuda_execute, var)

        assert np.allclose(val[0], [0.05127053])

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    @pytest.mark.benchmark(group="IntegratorMechanism")
    @pytest.mark.parametrize('mode', ['Python',
                                      pytest.param('LLVM', marks=pytest.mark.llvm),
                                      pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda])])
    def test_FitzHughNagumo_simple_vector(self, benchmark, mode):
        var = [1.0, 3.0]
        I = IntegratorMechanism(name="I",
                                default_variable=var,
                                function=FitzHughNagumoIntegrator)

        if mode == 'Python':
            val = I.execute(var)
            # LLVM versions report values of output ports. Collect it here
            val = [x.value for x in I.output_ports]
            benchmark(I.execute, var)
        elif mode == 'LLVM':
            e = pnlvm.execution.MechExecution(I)
            val = e.execute(var)
            benchmark(e.execute, var)
        elif mode == 'PTX':
            e = pnlvm.execution.MechExecution(I)
            val = e.cuda_execute(var)
            benchmark(e.cuda_execute, var)

        assert np.allclose(val[0], [0.05127053, 0.15379818])

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    @pytest.mark.benchmark(group="IntegratorMechanism")
    @pytest.mark.parametrize('mode', ['Python',
                                      pytest.param('LLVM', marks=pytest.mark.llvm),
                                      pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda])])
    def test_transfer_integrator(self, benchmark, mode):
        I = IntegratorMechanism(
            default_variable=[0 for i in range(VECTOR_SIZE)],
            function=Linear(slope=5.0))
        if mode == 'Python':
            val = benchmark(I.execute, [1.0 for i in range(VECTOR_SIZE)])
        elif mode == 'LLVM':
            e = pnlvm.execution.MechExecution(I)
            val = benchmark(e.execute, [1.0 for i in range(VECTOR_SIZE)])
        elif mode == 'PTX':
            e = pnlvm.execution.MechExecution(I)
            val = benchmark(e.cuda_execute, [1.0 for i in range(VECTOR_SIZE)])

        assert np.allclose(val, [[5.0 for i in range(VECTOR_SIZE)]])

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_accumulator_integrator(self):
        I = IntegratorMechanism(
            function=AccumulatorIntegrator(
                initializer=10.0,
                increment=15.0,
            )
        )
        # P = Process(pathway=[I])
        # accumulator integrator does not use input value (variable)

        # step 1:
        val = I.execute(20000)
        # value = 10 + 5
        # adjusted_value = 15 + 10
        # previous_value = 25
        # RETURN 25

        # step 2:
        val2 = I.execute(70000)
        # value = 25 + 5
        # adjusted_value = 30 + 10
        # previous_value = 30
        # RETURN 40
        assert (val, val2) == (25, 40)

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_adaptive_integrator(self):
        I = IntegratorMechanism(
            function=AdaptiveIntegrator(
                initializer=10.0,
                rate=0.5,
                offset=10,
            )
        )
        # P = Process(pathway=[I])
        # 10*0.5 + 1*0.5 + 10
        val = I.execute(1)
        assert val == 15.5

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_drift_diffusion_integrator(self):
        I = IntegratorMechanism(
            function=DriftDiffusionIntegrator(
                initializer=10.0,
                rate=10,
                time_step_size=0.5,
                offset=10,
            )
        )
        # P = Process(pathway=[I])
        # 10 + 10*0.5 + 0 + 10 = 25
        val = I.execute(1)
        assert np.allclose([[[25.]], [[0.5]]], val)

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_ornstein_uhlenbeck_integrator(self):
        I = IntegratorMechanism(
            function=OrnsteinUhlenbeckIntegrator(
                decay=0.5,
                initializer=10.0,
                rate=0.25,
                time_step_size=0.5,
                noise = 0.0,
                offset= 1.0
            )
        )
        # P = Process(pathway=[I])
        # value = previous_value + decay * (previous_value -  rate * new_value) * time_step_size + np.sqrt(
        # time_step_size * noise) * np.random.normal()
        # step 1:

        val = I.execute(1)
        # value = 10 + 0.5 * ( 10.0 - 0.25*1.0) * 0.5 + sqrt(0.25*0)*random_sample
        #       = 10 + 0.5*9.75*0.5
        #       = 12.4375
        # adjusted_value = 12.4375 + 1.0
        # previous_value = 13.4375
        # RETURN 13.4375

        # step 2:
        val2 = I.execute(1)
        # value = 13.4375 + 0.5 * ( 13.4375 - 0.25*1.0) * 0.5
        #       = 13.4375 + 3.296875
        # adjusted_value = 16.734375 + 1.0
        # previous_value = 17.734375
        # RETURN 31

# COMMENTED OUT UNTIL OU INTEGRATOR IS VALIDATED
        # assert (val, val2) == (13.4375, 17.734375)

# COMMENTED OUT UNTIL OU INTEGRATOR IS VALIDATED
    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_ornstein_uhlenbeck_integrator_time(self):
        OU = IntegratorMechanism(
            function=OrnsteinUhlenbeckIntegrator(
                initializer=10.0,
                rate=10,
                time_step_size=0.2,
                starting_point=0.5,
                decay=0.1,
                offset=10,
            )
        )
        time_0 = OU.function.previous_time  # t_0  = 0.5
        # np.testing.assert_allclose(time_0, [0.5], atol=1e-08)

        OU.execute(10)
        time_1 = OU.function.previous_time  # t_1  = 0.5 + 0.2 = 0.7
        # np.testing.assert_allclose(time_1, [0.7], atol=1e-08)

        for i in range(11):  # t_11 = 0.7 + 10*0.2 = 2.7
            OU.execute(10)
        time_12 = OU.function.previous_time # t_12 = 2.7 + 0.2 = 2.9
        # np.testing.assert_allclose(time_12, [2.9], atol=1e-08)

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    @pytest.mark.benchmark(group="IntegratorMechanism")
    @pytest.mark.parametrize('mode', ['Python',
                                      pytest.param('LLVM', marks=pytest.mark.llvm),
                                      pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda])])
    def test_integrator_no_function(self, benchmark, mode):
        I = IntegratorMechanism()
        if mode == 'Python':
            val = I.execute(10)
            benchmark(I.execute, 10)
        elif mode == 'LLVM':
            e = pnlvm.execution.MechExecution(I)
            val = e.execute([10])
            benchmark(e.execute, [10])
        elif mode == 'PTX':
            e = pnlvm.execution.MechExecution(I)
            val = e.cuda_execute([10])
            benchmark(e.cuda_execute, [10])

        assert np.allclose(val, [[5.0]])

class TestIntegratorInputs:
    # Part 1: VALID INPUT:

    # input = float

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_input_float(self):
        I = IntegratorMechanism(
            name='IntegratorMechanism',
            function=SimpleIntegrator(
            )
        )
        # P = Process(pathway=[I])
        val = float(I.execute(10.0))
        assert val == 10.0

    # input = list of length 1

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_input_list(self):
        I = IntegratorMechanism(
            name='IntegratorMechanism',
            function=SimpleIntegrator(
            )
        )
        # P = Process(pathway=[I])
        val = float(I.execute([10.0]))
        assert val == 10.0

    # input = list of length 5

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_input_list_len_5(self):

        I = IntegratorMechanism(
            name='IntegratorMechanism',
            default_variable=[0, 0, 0, 0, 0],
            function=SimpleIntegrator(
            )
        )
        # P = Process(pathway=[I])
        val = I.execute([10.0, 5.0, 2.0, 1.0, 0.0])[0]
        expected_output = [10.0, 5.0, 2.0, 1.0, 0.0]

        for i in range(len(expected_output)):
            v = val[i]
            e = expected_output[i]
            np.testing.assert_allclose(v, e, atol=1e-08, err_msg='Failed on expected_output[{0}]'.format(i))

    # input = numpy array of length 5

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_input_array_len_5(self):

        I = IntegratorMechanism(
            name='IntegratorMechanism',
            default_variable=[0, 0, 0, 0, 0],
            function=SimpleIntegrator(
            )
        )
        # P = Process(pathway=[I])
        input_array = np.array([10.0, 5.0, 2.0, 1.0, 0.0])
        val = I.execute(input_array)[0]
        expected_output = [10.0, 5.0, 2.0, 1.0, 0.0]

        for i in range(len(expected_output)):
            v = val[i]
            e = expected_output[i]
            np.testing.assert_allclose(v, e, atol=1e-08, err_msg='Failed on expected_output[{0}]'.format(i))

    # Part 2: INVALID INPUT

    # input = list of length > default length

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_input_array_greater_than_default(self):

        with pytest.raises(MechanismError) as error_text:
            I = IntegratorMechanism(
                name='IntegratorMechanism',
                default_variable=[0, 0, 0]
            )
            # P = Process(pathway=[I])
            I.execute([10.0, 5.0, 2.0, 1.0, 0.0])
        assert "does not match required length" in str(error_text.value)

    # input = list of length < default length

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_input_array_less_than_default(self):

        with pytest.raises(MechanismError) as error_text:
            I = IntegratorMechanism(
                name='IntegratorMechanism',
                default_variable=[0, 0, 0, 0, 0]
            )
            # P = Process(pathway=[I])
            I.execute([10.0, 5.0, 2.0])
        assert "does not match required length" in str(error_text.value)


# ======================================= RATE TESTS ============================================
class TestIntegratorRate:
    # VALID RATE:

    # rate = float, integration_type = simple

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_type_simple_rate_float(self):
        I = IntegratorMechanism(
            name='IntegratorMechanism',
            function=SimpleIntegrator(
                rate=5.0
            )
        )
        # P = Process(pathway=[I])
        val = float(I.execute(10.0))
        assert val == 50.0

    # rate = float, increment = float, integration_type = accumulator

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_type_accumulator_rate_and_increment_float(self):
        I = IntegratorMechanism(
            name='IntegratorMechanism',
            function=AccumulatorIntegrator(
                rate=2.0,
                increment=3.0
            )
        )
        # P = Process(pathway=[I])
        float(I.execute())
        val = float(I.execute())
        assert val == 9.0

    # rate = float, integration_type = diffusion

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_type_diffusion_rate_float(self):
        I = IntegratorMechanism(
            name='IntegratorMechanism',
            function=DriftDiffusionIntegrator(
                rate=5.0
            )
        )
        # P = Process(pathway=[I])
        val = I.execute(10.0)
        assert np.allclose([[[50.0]], [[1.0]]], val)

    # rate = list, integration_type = simple

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_type_simple_rate_list(self):
        I = IntegratorMechanism(
            name='IntegratorMechanism',
            default_variable=[0, 0, 0],
            function=SimpleIntegrator(
                rate=[5.0, 5.0, 5.0]
            )
        )
        # P = Process(pathway=[I])
        val = list(I.execute([10.0, 10.0, 10.0])[0])
        assert val == [50.0, 50.0, 50.0]

    # rate = float, increment = list, integration_type = accumulator

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_type_accumulator_rate_float_increment_list(self):
        I = IntegratorMechanism(
            default_variable=[0, 0, 0],
            name='IntegratorMechanism',
            function=AccumulatorIntegrator(
                rate = 2.0,
                increment=[4.0, 5.0, 6.0]
            )
        )
        # P = Process(pathway=[I])
        list(I.execute([10.0, 10.0, 10.0])[0])
        val = list(I.execute([10.0, 10.0, 10.0])[0])
        assert val == [12.0, 15.0, 18.0]

    # rate = float, increment = list, integration_type = accumulator

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_type_accumulator_rate_list_increment_float(self):
        I = IntegratorMechanism(
            default_variable=[0, 0, 0],
            name='IntegratorMechanism',
            function=AccumulatorIntegrator(
                rate = [2.0, 3.0, 4.0],
                increment=5.0
            )
        )
        # P = Process(pathway=[I])
        list(I.execute([10.0, 10.0, 10.0])[0])
        val = list(I.execute([10.0, 10.0, 10.0])[0])
        assert val == [15.0, 20.0, 25.0]

    # rate = list, increment = list, integration_type = accumulator

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_type_accumulator_rate_and_increment_list(self):
        I = IntegratorMechanism(
            default_variable=[0, 0, 0],
            name='IntegratorMechanism',
            function=AccumulatorIntegrator(
                rate = [1.0, 2.0, 3.0],
                increment=[4.0, 5.0, 6.0]
            )
        )
        # P = Process(pathway=[I])
        list(I.execute([10.0, 10.0, 10.0])[0])
        val = list(I.execute([10.0, 10.0, 10.0])[0])
        assert val == [8.0, 15.0, 24.0]


    # rate = list, integration_type = diffusion

    # def test_integrator_type_diffusion_rate_list(self):
    #     I = IntegratorMechanism(
    #         default_variable=[0, 0, 0],
    #         name='IntegratorMechanism',
    #         function=DriftDiffusionIntegrator(
    #             rate=[5.0, 5.0, 5.0],
    #             threshold=[100.0, 100.0, 100.0]
    #         )
    #     )
    #     # P = Process(pathway=[I])
    #     val = list(I.execute([10.0, 10.0, 10.0])[0])
    #     assert val == [50.0, 50.0, 50.0]

    # rate = list, integration_type = diffusion

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_type_adaptive_rate_list(self):
        I = IntegratorMechanism(
            default_variable=[0, 0, 0],
            name='IntegratorMechanism',
            function=AdaptiveIntegrator(
                rate=[0.5, 0.5, 0.5]
            )
        )
        # P = Process(pathway=[I])
        val = list(I.execute([10.0, 10.0, 10.0])[0])
        assert val == [5.0, 5.0, 5.0]

    # rate = float, integration_type = modulatory

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_type_adaptive_rate_float_input_list(self):
        I = IntegratorMechanism(
            default_variable=[0, 0, 0],
            name='IntegratorMechanism',
            function=AdaptiveIntegrator(
                rate=0.5
            )
        )
        # P = Process(pathway=[I])
        val = list(I.execute([10.0, 10.0, 10.0])[0])
        assert val == [5.0, 5.0, 5.0]

    # rate = float, integration_type = modulatory

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_type_adaptive_rate_float(self):
        I = IntegratorMechanism(
            name='IntegratorMechanism',
            function=AdaptiveIntegrator(
                rate=0.5
            )
        )
        # P = Process(pathway=[I])
        val = list(I.execute(10.0))
        assert val == [5.0]

    # INVALID RATE:

    # rate = list, execute float, integration_type = simple

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_type_adaptive_variable_and_rate_conflict(self):
        with pytest.raises(IntegratorMechanismError) as error_text:
            I = IntegratorMechanism(
                    name='IntegratorMechanism',
                    default_variable=[0],
                    function=AdaptiveIntegrator(rate=[0.5, 0.6])
            )
        error_msg_a = "Shape of 'variable' for function specified for IntegratorMechanism (AdaptiveIntegrator Function"
        error_msg_b = "(2,)) does not match the shape of the 'default_variable' specified for the 'Mechanism'."
        assert error_msg_a in str(error_text.value)
        assert error_msg_b in str(error_text.value)

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_type_simple_rate_list_input_float(self):
        with pytest.raises(MechanismError) as error_text:
            I = IntegratorMechanism(
                name='IntegratorMechanism',
                function=SimpleIntegrator(
                    rate=[5.0, 5.0, 5.0]
                )
            )
            result = I.execute(10.0)
            float(result)
        returned_error_msg = str(error_text.value)
        # Need to break up error message because length value returned ([10.0] vs. [ 10.0]) may differ by Python vers.
        error_msg_a1 = 'Length (1) of input'
        error_msg_a2 = 'does not match required length (3) for input to '
        error_msg_b = "InputPort 'InputPort-0' of IntegratorMechanism"
        assert [msg in returned_error_msg for msg in {error_msg_a1, error_msg_a2, '10'}]
        assert error_msg_b in str(error_text.value)

    # rate = list len 2, incrment = list len 3, integration_type = accumulator

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_type_accumulator_increment_list_input_float(self):
        with pytest.raises(FunctionError) as error_text:
            I = IntegratorMechanism(
                name='IntegratorMechanism',
                function=AccumulatorIntegrator(
                    rate=[1.0, 2.0],
                    increment=[3.0, 4.0, 5.0]
                ))

        error_msg_a = "The parameters with len>1 specified for AccumulatorIntegrator Function-0 "
        error_msg_b = "(['rate', 'increment']) don't all have the same length"
        assert error_msg_a in str(error_text.value)
        assert error_msg_b in str(error_text.value)


    # @pytest.mark.mechanism
    # @pytest.mark.integrator_mechanism
    # def test_integrator_type_diffusion_rate_list_input_float(self):
    #     with pytest.raises(FunctionError) as error_text:
    #         I = IntegratorMechanism(
    #             name='IntegratorMechanism',
    #             function=DriftDiffusionIntegrator(
    #                 rate=[5.0, 5.0, 5.0]
    #             )
    #         )
    #         # P = Process(pathway=[I])
    #         float(I.execute(10.0))
    #     assert "array specified for the rate parameter" in str(error_text.value)
    #     assert "must match the length" in str(error_text.value)
    #     assert "of the default input" in str(error_text.value)

    # @pytest.mark.mechanism
    # @pytest.mark.integrator_mechanism
    # def test_accumulator_integrator(self):
    #     I = IntegratorMechanism(
    #             function = AccumulatorIntegrator(
    #                 initializer = 10.0,
    #                 rate = 5.0,
    #                 increment= 1.0
    #             )
    #         )
    # #     P = Process(pathway=[I])

    #     # value = previous_value * rate + noise + increment
    #     # step 1:
    #     val = I.execute()
    #     # value = 10.0 * 5.0 + 0 + 1.0
    #     # RETURN 51

    #     # step 2:
    #     val2 = I.execute(2000)
    #     # value = 51*5 + 0 + 1.0
    #     # RETURN 256
    #     assert (val, val2) == (51, 256)


# ======================================= NOISE TESTS ============================================
class TestIntegratorNoise:

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_simple_noise_fn(self):
        I = IntegratorMechanism(
            name='IntegratorMechanism',
            function=SimpleIntegrator(
                noise=NormalDist()
            ),
        )

        val = float(I.execute(10))

        I.function.reset(5.0)

        val2 = float(I.execute(0))

        np.testing.assert_allclose(val, 9.306822898004032)
        np.testing.assert_allclose(val2, 6.000180029830552)

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_simple_noise_fn_var_list(self):
        I = IntegratorMechanism(
            name='IntegratorMechanism',
            default_variable=[0, 0, 0, 0],
            function=SimpleIntegrator(
                noise=NormalDist(),
            ),
        )

        val = I.execute([10, 10, 10, 10])[0]

        np.testing.assert_allclose(val, [10.66053518, 11.10887925, 9.0840107, 10.30157835])

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_accumulator_noise_fn(self):
        I = IntegratorMechanism(
            name='IntegratorMechanism',
            function=AccumulatorIntegrator(
                noise=NormalDist()
            ),
        )

        val = float(I.execute(10))

        np.testing.assert_allclose(val, -0.6931771019959673)

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_accumulator_noise_fn_var_list(self):
        I = IntegratorMechanism(
            name='IntegratorMechanism',
            default_variable=[0, 0, 0, 0],
            function=AccumulatorIntegrator(
                noise=NormalDist(),
            ),
        )

        val = I.execute([10, 10, 10, 10])[0]
        np.testing.assert_allclose(val, [0.66053518, 1.10887925, -0.9159893, 0.30157835])

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_adaptive_noise_fn(self):
        I = IntegratorMechanism(
            name='IntegratorMechanism',
            function=AdaptiveIntegrator(
                noise=NormalDist()
            ),
        )

        val = float(I.execute(10))

        np.testing.assert_allclose(val, 9.306822898004032)

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_adaptive_noise_fn_var_list(self):
        I = IntegratorMechanism(
            name='IntegratorMechanism',
            default_variable=[0, 0, 0, 0],
            function=AdaptiveIntegrator(
                noise=NormalDist(),
            ),
        )

        val = I.execute([10, 10, 10, 10])[0]

        np.testing.assert_allclose(val, [10.66053518, 11.10887925, 9.0840107, 10.30157835])

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_drift_diffusion_noise_val(self):
        I = IntegratorMechanism(
            name='IntegratorMechanism',
            function=DriftDiffusionIntegrator(
                noise=5.0,
            ),
        )

        val = I.execute(10.0)
        assert np.allclose(val, [[[4.29013944]], [[ 1.        ]]])

# COMMENTED OUT UNTIL OU INTEGRATOR IS VALIDATED
    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_integrator_ornstein_uhlenbeck_noise_val(self):
        I = IntegratorMechanism(
            name='IntegratorMechanism',
            function=OrnsteinUhlenbeckIntegrator(
                noise=2.0,
                decay=0.5,
                initializer=1.0,
                rate=0.25
            ),
        )

        # val = 1.0 + 0.5 * (1.0 - 0.25 * 2.5) * 1.0 + np.sqrt(1.0 * 2.0) * np.random.normal()


        val = I.execute(2.5)

        # np.testing.assert_allclose(val, 4.356601554140335)

class TestStatefulness:

    def test_has_initializers(self):
        I = IntegratorMechanism()
        assert I.has_initializers
        assert hasattr(I, "reset_stateful_function_when")

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    @pytest.mark.benchmark(group="IntegratorMechanism")
    @pytest.mark.parametrize('mode', ['Python',
                                      pytest.param('LLVMExec', marks=pytest.mark.llvm),
                                      pytest.param('LLVMRun', marks=pytest.mark.llvm),
                                      pytest.param('PTXExec', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                      pytest.param('PTXRun', marks=[pytest.mark.llvm, pytest.mark.cuda])])
    @pytest.mark.parametrize('cond0, cond1, expected', [
        (pnl.Never(), pnl.AtTrial(2),
         [[np.array([0.5]), np.array([0.5])],
          [np.array([0.75]), np.array([0.75])],
          [np.array([0.875]), np.array([0.5])],   # I2 resets at Trial 2
          [np.array([0.9375]), np.array([0.75])],
          [np.array([0.96875]), np.array([0.875])],
          [np.array([0.984375]), np.array([0.9375])],
          [np.array([0.9921875]), np.array([0.96875])]]),
        (pnl.Never(), pnl.AtTrialStart(),
         [[np.array([0.5]), np.array([0.5])],
          [np.array([0.75]), np.array([0.5])],
          [np.array([0.875]), np.array([0.5])],
          [np.array([0.9375]), np.array([0.5])],
          [np.array([0.96875]), np.array([0.5])],
          [np.array([0.984375]), np.array([0.5])],
          [np.array([0.9921875]), np.array([0.5])]]),
        (pnl.AtPass(0), pnl.AtTrial(2),
         [[np.array([0.5]), np.array([0.5])],
          [np.array([0.5]), np.array([0.75])],
          [np.array([0.5]), np.array([0.5])],   # I2 resets at Trial 2
          [np.array([0.5]), np.array([0.75])],
          [np.array([0.5]), np.array([0.875])],
          [np.array([0.5]), np.array([0.9375])],
          [np.array([0.5]), np.array([0.96875])]]),
        ], ids=lambda x: str(x) if isinstance(x, pnl.Condition) else "")
    def test_reset_stateful_function_when_composition(self, mode, cond0, cond1, expected):
        I1 = pnl.IntegratorMechanism()
        I2 = pnl.IntegratorMechanism()
        I1.reset_stateful_function_when = cond0
        I2.reset_stateful_function_when = cond1
        C = pnl.Composition()
        C.add_node(I1)
        C.add_node(I2)

        C.run(inputs={I1: [[1.0]], I2: [[1.0]]}, num_trials=7, bin_execute=mode)

        assert np.allclose(expected, C.results)

    def test_reset_stateful_function_when(self):
        I1 = IntegratorMechanism()
        I2 = IntegratorMechanism()
        I2.reset_stateful_function_when = AtTrial(2)
        C = Composition(pathways=[[I1], [I2]])

        C.run(inputs={I1: [[1.0]],
                      I2: [[1.0]]},
              num_trials=7,
              reset_stateful_functions_when=AtTrial(3))

        expected_results = [[np.array([0.5]), np.array([0.5])],
                            [np.array([0.75]), np.array([0.75])],
                            [np.array([0.875]), np.array([0.5])],   # I2 resets at Trial 2
                            [np.array([0.5]), np.array([0.75])],    # I1 resets at Trial 3
                            [np.array([0.75]), np.array([0.875])],
                            [np.array([0.875]), np.array([0.9375])],
                            [np.array([0.9375]), np.array([0.96875])]]

        assert np.allclose(expected_results, C.results)


class TestDualAdaptiveIntegrator:

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_utility_integrator_default(self):
        # default params:
        # initial_short_term_avg = 0.0
        # initial_long_term_avg = 0.0
        # short_term_rate = 1.0
        # long_term_rate = 1.0

        U = IntegratorMechanism(
            name = "DualAdaptiveIntegrator",
            function=DualAdaptiveIntegrator(
            )

        )

        engagement = []
        short_term_util = []
        long_term_util = []
        for i in range(50):
            engagement.append(U.execute([1])[0][0])
            short_term_util.append(U.function.short_term_logistic[0])
            long_term_util.append(U.function.long_term_logistic[0])
        print("engagement = ", engagement)
        print("short_term_util = ", short_term_util)
        print("long_term_util = ", long_term_util)

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_utility_integrator_short_minus_long(self):
        # default params:
        # initial_short_term_avg = 0.0
        # initial_long_term_avg = 0.0
        # short_term_rate = 1.0
        # long_term_rate = 1.0

        U = IntegratorMechanism(
            name = "DualAdaptiveIntegrator",
            function=DualAdaptiveIntegrator(
                operation=pnl.S_MINUS_L
            )

        )

        engagement = []
        short_term_util = []
        long_term_util = []
        for i in range(50):
            engagement.append(U.execute([1])[0][0])
            short_term_util.append(U.function.short_term_logistic[0])
            long_term_util.append(U.function.long_term_logistic[0])
        print("engagement = ", engagement)
        print("short_term_util = ", short_term_util)
        print("long_term_util = ", long_term_util)

    @pytest.mark.mechanism
    @pytest.mark.integrator_mechanism
    def test_utility_integrator_short_plus_long(self):
        # default params:
        # initial_short_term_avg = 0.0
        # initial_long_term_avg = 0.0
        # short_term_rate = 1.0
        # long_term_rate = 1.0

        U = IntegratorMechanism(
            name = "DualAdaptiveIntegrator",
            function=DualAdaptiveIntegrator(
                operation=pnl.SUM
            )

        )

        engagement = []
        short_term_util = []
        long_term_util = []
        for i in range(50):
            engagement.append(U.execute([1])[0][0])
            short_term_util.append(U.function.short_term_logistic[0])
            long_term_util.append(U.function.long_term_logistic[0])
        print("engagement = ", engagement)
        print("short_term_util = ", short_term_util)
        print("long_term_util = ", long_term_util)

    # @pytest.mark.mechanism
    # @pytest.mark.integrator_mechanism
    # def test_plot_utility_integrator(self):
        # from matplotlib import pyplot as plt
        # from mpl_toolkits.mplot3d import Axes3D
        # from matplotlib import cm
        # import numpy as np
        # fig = plt.figure()
        # ax = fig.gca(projection='3d')
        #
        # def logistic(variable):
        #
        #     try:
        #         return_val = 1 / (1 + np.exp(-variable))
        #     except (Warning):
        #         # handle RuntimeWarning: overflow in exp
        #         return_val = 0
        #
        #     return return_val
        #
        # short = np.linspace(0, 1, 20)
        # long = np.linspace(0, 1, 20)
        # short_grid, long_grid = np.meshgrid(short, long, sparse=True)
        #
        # short_logistic = logistic(short)
        # long_logistic = logistic(long)
        # short_grid_logistic, long_grid_logistic = np.meshgrid(short_grid, long_grid, sparse=True)
        #
        # z = (1-short_grid_logistic)*long_grid_logistic
        # surf = ax.plot_surface(short_grid_logistic, long_grid_logistic, z,
        #                        cmap=cm.gray,
        #                        linewidth=0,
        #                        # antialiased=False,
        #
        #                        )
        # # fig.colorbar(surf, shrink=0.5, aspect=5)
        # plt.show()

        # def logistic(x):
        #
        #     try:
        #         return_val = 1 / (1 + np.exp(-x))
        #     except (Warning):
        #         # handle RuntimeWarning: overflow in exp
        #         return_val = 0
        #
        #     return return_val
        #
        # def combine(s, l):
        #     return (1-s)*l
        #
        # import numpy as np
        # from mpl_toolkits.mplot3d import Axes3D
        # import matplotlib.pyplot as plt
        #
        #
        # fig = plt.figure()
        # ax = fig.add_subplot(111, projection='3d')
        # s = l = np.arange(0, 1.0, 0.05)
        # s = list(map(logistic, s))
        # l = list(map(logistic, l))
        # S, L = np.meshgrid(s, l)
        # zs = np.array([combine(s, l) for s, l in zip(np.ravel(S), np.ravel(L))])
        # Z = zs.reshape(S.shape)
        #
        # ax.plot_surface(S, L, Z)
        #
        # ax.set_xlabel('short term')
        # ax.set_ylabel('long term')
        # ax.set_zlabel('engagement')
        #
        # plt.show()


    # @pytest.mark.mechanism
    # @pytest.mark.integrator_mechanism
    # def test_FitzHughNagumo_gilzenrat(self):
    #
    #     F = IntegratorMechanism(
    #         name='IntegratorMech-FitzHughNagumoFunction',
    #         function=FitzHughNagumoIntegrator(
    #             time_step_size=0.1,
    #             initial_v=0.2,
    #             initial_w=0.0,
    #             t_0=0.0,
    #             time_accumulator_v=1.0,
    #             a_v=-1.0,
    #             b_v=1.5,
    #             c_v=-0.5,
    #             d_v=0.0,
    #             e_v=-1.0,
    #             f_v=0.0,
    #             time_accumulator_w=100.0,
    #             a_w=1.0,
    #             b_w=-0.5,
    #             c_w=0.0
    #         )
    #     )
    #     plot_v_list = []
    #     plot_w_list = []
    #
    #     expected_v_list = []
    #     expected_w_list = []
    #     stimulus = 0.0
    #     for i in range(10):
    #
    #         for j in range(50):
    #             new_v = F.execute(stimulus)[0][0]
    #             new_w = F.execute(stimulus)[1][0]
    #             # ** uncomment the lines below if you want to view the plot:
    #             plot_v_list.append(new_v)
    #             plot_w_list.append(new_w)
    #         expected_v_list.append(new_v)
    #         expected_w_list.append(new_w)
    #     # print(plot_v_list)
    #     # print(plot_w_list)
    #     # ** uncomment the lines below if you want to view the plot:
    #     import matplotlib.pyplot as plt
    #     plt.plot(plot_v_list)
    #     plt.plot(plot_w_list)
    #     plt.show()
    #
    #     # np.testing.assert_allclose(expected_v_list, [1.9861589924245777, 1.9184159304279109, 1.7920107368145777,
    #     #                                              1.6651158106802393, 1.5360917598075965, 1.4019128309448776,
    #     #                                              1.2568420252868404, 1.08773745582042, 0.8541804646541804,
    #     #                                              0.34785588139530099])
    #     # np.testing.assert_allclose(expected_w_list, [0.28713219302304327, 0.65355262255707869, 0.9581082373550347,
    #     #                                              1.2070585850028435, 1.4068978270680454, 1.5629844531368104,
    #     #                                              1.6793901854329185, 1.7583410650743645, 1.7981128658110572,
    #     #                                              1.7817328532815251])
    #     #
    #
    # def test_FitzHughNagumo_gilzenrat_low_electrotonic_coupling(self):
    #
    #     F = IntegratorMechanism(
    #         name='IntegratorMech-FitzHughNagumoFunction',
    #         function=FitzHughNagumoIntegrator(
    #             time_step_size=0.1,
    #             initial_v=0.2,
    #             initial_w=0.0,
    #             t_0=0.0,
    #             time_accumulator_v=1.0,
    #             a_v=-1.0,
    #             b_v=0.5,
    #             c_v=0.5,
    #             d_v=0.0,
    #             e_v=-1.0,
    #             f_v=0.0,
    #             electrotonic_coupling=0.55,
    #             time_accumulator_w=100.0,
    #             a_w=1.0,
    #             b_w=-0.5,
    #             c_w=0.0
    #         )
    #     )
    #     plot_v_list = []
    #     plot_w_list = []
    #
    #     expected_v_list = []
    #     expected_w_list = []
    #     stimulus = 0.0
    #     for i in range(10):
    #
    #         for j in range(600):
    #             new_v = F.execute(stimulus)[0][0]
    #             new_w = F.execute(stimulus)[1][0]
    #             # ** uncomment the lines below if you want to view the plot:
    #             plot_v_list.append(new_v)
    #             plot_w_list.append(new_w)
    #         expected_v_list.append(new_v)
    #         expected_w_list.append(new_w)
    #     # print(plot_v_list)
    #     # print(plot_w_list)
    #     # ** uncomment the lines below if you want to view the plot:
    #     import matplotlib.pyplot as plt
    #     plt.plot(plot_v_list)
    #     plt.plot(plot_w_list)
    #     plt.show()
    #
    #     # np.testing.assert_allclose(expected_v_list, [1.9861589924245777, 1.9184159304279109, 1.7920107368145777,
    #     #                                              1.6651158106802393, 1.5360917598075965, 1.4019128309448776,
    #     #                                              1.2568420252868404, 1.08773745582042, 0.8541804646541804,
    #     #                                              0.34785588139530099])
    #     # np.testing.assert_allclose(expected_w_list, [0.28713219302304327, 0.65355262255707869, 0.9581082373550347,
    #     #                                              1.2070585850028435, 1.4068978270680454, 1.5629844531368104,
    #     #                                              1.6793901854329185, 1.7583410650743645, 1.7981128658110572,
    #     #                                              1.7817328532815251])
    #     #



#
# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
#
# *********************************************  INTERFACE FUNCTIONS ***************************************************

"""

* InterfacePortMap

"""

import numpy as np
import typecheck as tc
import warnings

from psyneulink.core.components.functions.function import Function_Base
from psyneulink.core.globals.context import ContextFlags
from psyneulink.core.globals.keywords import \
    FUNCTION_OUTPUT_TYPE_CONVERSION, PARAMETER_PORT_PARAMS, PORT_MAP_FUNCTION, TRANSFER_FUNCTION_TYPE, \
    PREFERENCE_SET_NAME
from psyneulink.core.globals.parameters import Parameter
from psyneulink.core.globals.preferences.basepreferenceset import \
    PreferenceEntry, PreferenceLevel, is_pref_set, REPORT_OUTPUT_PREF
from psyneulink.core.globals.utilities import convert_to_np_array


__all__ = ['InterfaceFunction', 'InterfacePortMap']

class InterfaceFunction(Function_Base):
    """Simple functions for CompositionInterfaceMechanisms
    """
    componentType = TRANSFER_FUNCTION_TYPE


class InterfacePortMap(InterfaceFunction):
    """
    Identity(                \
             default_variable, \
             params=None,      \
             owner=None,       \
             name=None,        \
             prefs=None        \
            )

    .. _Identity:

    Returns `variable <InterfacePortMap.variable>`.

    Arguments
    ---------

    default_variable : number or np.array : default class_defaults.variable
        specifies a template for the value to be transformed.

    params : Dict[param keyword: param value] : default None
        a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the
        function.  Values specified for parameters in the dictionary override any assigned to those parameters in
        arguments of the constructor.

    owner : Component
        `component <Component>` to which to assign the Function.

    name : str : default see `name <Function.name>`
        specifies the name of the Function.

    prefs : PreferenceSet or specification dict : default Function.classPreferences
        specifies the `PreferenceSet` for the Function (see `prefs <Function_Base.prefs>` for details).

    Attributes
    ----------

    variable : number or np.array
        contains value to be transformed.

    owner : Component
        `component <Component>` to which the Function has been assigned.

    name : str
        the name of the Function; if it is not specified in the **name** argument of the constructor, a default is
        assigned by FunctionRegistry (see `Registry_Naming` for conventions used for default and duplicate names).

    prefs : PreferenceSet or specification dict : Function.classPreferences
        the `PreferenceSet` for function; if it is not specified in the **prefs** argument of the Function's
        constructor, a default is assigned using `classPreferences` defined in __init__.py (see `Preferences`
        for details).
    """

    componentName = PORT_MAP_FUNCTION

    classPreferences = {
        PREFERENCE_SET_NAME: 'LinearClassPreferences',
        REPORT_OUTPUT_PREF: PreferenceEntry(False, PreferenceLevel.INSTANCE),
    }

    class Parameters(InterfaceFunction.Parameters):
        """
            Attributes
            ----------

                corresponding_input_port
                    see `corresponding_input_port <InterfacePortMap.corresponding_input_port>`

                    :default value: None
                    :type:
        """
        corresponding_input_port = Parameter(
            None,
            structural=True,
            stateful=False,
            loggable=False
        )

    @tc.typecheck
    def __init__(self,
                 default_variable=None,
                 corresponding_input_port=None,
                 params=None,
                 owner=None,
                 prefs: tc.optional(is_pref_set) = None):

        super().__init__(
            default_variable=default_variable,
            corresponding_input_port=corresponding_input_port,
            params=params,
            owner=owner,
            prefs=prefs,
        )

        # self.functionOutputType = None

    def _function(
        self,
        variable=None,
        context=None,
        params=None,

    ):
        """
        Return: The item of `value <InterfacePortMap.value>` whose index corresponds to the index of
        `corresponding_input_port <InterfacePortMap.corresponding_input_port>` in `input_ports
        <InterfacePortMap.input_ports>`

        Arguments
        ---------

        variable : number or np.array : default class_defaults.variable
           a single value or array to be transformed.

        corresponding_input_port : InputPort : default None
            the InputPort on the owner CompositionInterfaceMechanism to which this OutputPort corresponds

        params : Dict[param keyword: param value] : default None
            a `parameter dictionary <ParameterPort_Specification>` that specifies the parameters for the
            function.  Values specified for parameters in the dictionary override any assigned to those parameters in
            arguments of the constructor.


        Returns
        -------

        The item of `value <InterfacePortMap.value>` whose index corresponds to the index of
        `corresponding_input_port <InterfacePortMap.corresponding_input_port>` in `input_ports
        <InterfacePortMap.input_ports>`

        """
        index = self.corresponding_input_port.position_in_mechanism

        if self.corresponding_input_port.owner.parameters.value._get(context) is not None:

            # If CIM's variable does not match its value, then a new pair of ports was added since the last execution
            input_values = convert_to_np_array(
                self.corresponding_input_port.owner.get_input_values(context)
            )
            if not np.shape(input_values) == np.shape(self.corresponding_input_port.owner.parameters.value._get(context)):
                return self.corresponding_input_port.owner.defaults.variable[index]

            # If the variable is 1D (e.g. [0. , 0.], NOT [[0. , 0.]]), and the index is 0, then return whole variable
            # np.atleast_2d fails in cases like var = [[0., 0.], [0.]] (transforms it to [[[0., 0.], [0.]]])
            if index == 0:
                if not isinstance(variable[0], (list, np.ndarray)):
                    return variable
            return variable[index]
        # CIM value = None, use CIM's default variable instead
        return self.corresponding_input_port.owner.defaults.variable[index]

    def _get_input_struct_type(self, ctx):
        #FIXME: Workaround for CompositionInterfaceMechanism that
        #       does not update its default shape
        from psyneulink import CompositionInterfaceMechanism
        if hasattr(self.owner, 'owner') and isinstance(self.owner.owner, CompositionInterfaceMechanism):
            warnings.warn("Shape mismatch: {} input: {} vs. ({}, MV:{})".format(
                self, self.defaults.variable, self.owner._variable_spec,
                self.owner.owner.function.defaults.value))
            return ctx.get_output_struct_type(self.owner.owner.function)
        return ctx.get_input_struct_type(super())

    def _get_output_struct_type(self, ctx):
        index = self.corresponding_input_port.position_in_mechanism
        input_type = ctx.get_input_struct_type(self)
        return input_type.elements[index]

    def _gen_llvm_function_body(self, ctx, builder, _1, _2, arg_in, arg_out, *, tags:frozenset):
        index = self.corresponding_input_port.position_in_mechanism
        val = builder.load(builder.gep(arg_in, [ctx.int32_ty(0), ctx.int32_ty(index)]))
        builder.store(val, arg_out)
        return builder



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ********************************************* Time ***************************************************************

"""

.. _Time_Overview:

Overview
--------

:doc:`Scheduler`\\ s maintain `Clock` objects to track time. The current time in \
relation to a :doc:`Scheduler` is stored in :class:`Scheduler.clock.time <Time>` \
or :class:`Scheduler.clock.simple_time <SimpleTime>`

"""

import copy
import enum
import functools
import types

__all__ = [
    'Clock', 'TimeScale', 'Time', 'SimpleTime', 'TimeHistoryTree', 'TimeScaleError'
]


class TimeScaleError(Exception):

    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)


# Time scale modes
@functools.total_ordering
class TimeScale(enum.Enum):
    """Represents divisions of time used by the `Scheduler`, `Conditions <Condition>`, and the **time_scale**
    argument of a Composition's `run <Composition.run>` method.

    The values of TimeScale are defined as follows (in order of increasingly coarse granularity):

    Attributes
    ----------

    TIME_STEP
        the nuclear unit of time, corresponding to the execution of all `Mechanism <Mechanism>`\\ s allowed to execute
        from a single `consideration_set <consideration_set>` of a `Scheduler`, and which are considered to have
        executed simultaneously.

    PASS
        a full iteration through all of the `consideration_sets <consideration_set>` in a `Scheduler's <Scheduler>`
        `consideration_queue`, consisting of one or more `TIME_STEPs <TIME_STEP>`, over which every `Component
        <Component>` `specified to a Scheduler <Scheduler_Creation>` is considered for execution at least once.

    TRIAL
        an open-ended unit of time consisting of all actions that occurs within the scope of a single input to a
        `Composition <Composition>`.

    RUN
        the scope of a call to the `run <Composition.run>` method of a `Composition <Composition>`,
        consisting of one or more `TRIALs <TimeScale.TRIAL>`.

    LIFE
        the scope of time since the creation of an object.
    """
    TIME_STEP = 0
    PASS = 1
    TRIAL = 2
    RUN = 3
    LIFE = 4

    # ordering based on enum.OrderedEnum example
    # https://docs.python.org/3/library/enum.html#orderedenum
    # https://stackoverflow.com/a/39269589/3131666
    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented

    @classmethod
    def get_parent(cls, time_scale):
        """
        Returns
        -------
            the TimeScale one level wider in scope than time_scale : :class:`TimeScale`
        """
        return cls(time_scale.value + 1)

    @classmethod
    def get_child(cls, time_scale):
        """
        Returns
        -------
            the TimeScale one level smaller in scope than time_scale : :class:`TimeScale`
        """
        return cls(time_scale.value - 1)


class Clock:
    """
    Stores a history of :class:`TimeScale`\\ s that have occurred, and keep track of a \
    current `Time`. Used in relation to a :doc:`Scheduler`

    Attributes
    ----------
        history : `TimeHistoryTree`
            a root `TimeHistoryTree` associated with this Clock
    """
    def __init__(self):
        self.history = TimeHistoryTree()
        self._simple_time = SimpleTime()

    def __repr__(self):
        return 'Clock({0})'.format(self.time.__repr__())

    def _increment_time(self, time_scale):
        """
        Calls `self.history.increment_time <TimeHistoryTree.increment_time>`
        """
        self.history.increment_time(time_scale)

    def get_total_times_relative(self, query_time_scale, base_time_scale, base_index=None):
        """
        Convenience simplified wrapper for `TimeHistoryTree.get_total_times_relative`

        Arguments
        ---------
            query_time_scale : :class:`TimeScale`
                the unit of time whose number of ticks to be returned

            base_time_scale : :class:`TimeScale`
                the unit of time over which the number of **query_time_scale** ticks
                should be returned

            base_index : int
                the **base_index**\\ th **base_time_scale** over which the number of
                **query_time_scale** ticks should be returned
        Returns
        -------
            the number of query_time_scale s that have occurred during the scope \
            of the base_index th base_time_scale : int
        """
        if base_index is None:
            base_index = self.get_time_by_time_scale(base_time_scale)

        return self.history.get_total_times_relative(
            query_time_scale,
            {base_time_scale: base_index}
        )

    def get_time_by_time_scale(self, time_scale):
        """
        Arguments
        ---------
            time_scale : :class:`TimeScale`

        Returns
        -------
            the current value of the time unit corresponding to time_scale \
            for this Clock : int
        """
        return self.time._get_by_time_scale(time_scale)

    @property
    def time(self):
        """
        the current time : `Time`
        """
        return self.history.current_time

    @property
    def simple_time(self):
        """
        the current time in simple format : `SimpleTime`
        """
        self._simple_time.run = self.time.run
        self._simple_time.trial = self.time.trial
        self._simple_time.time_step = self.time.time_step
        return self._simple_time

    @property
    def previous_time(self):
        """
        the time that has occurred last : `Time`
        """
        return self.history.previous_time


class Time(types.SimpleNamespace):
    """
    Represents an instance of time, having values for each :class:`TimeScale`

    Attributes
    ----------
        life : int : 0
            the `TimeScale.LIFE` value

        run : int : 0
            the `TimeScale.RUN` value

        trial : int : 0
            the `TimeScale.TRIAL` value

        pass_ : int : 0
            the `TimeScale.PASS` value

        time_step : int : 0
            the `TimeScale.TIME_STEP` value

    """
    _time_scale_attr_map = {
        TimeScale.TIME_STEP: 'time_step',
        TimeScale.PASS: 'pass_',
        TimeScale.TRIAL: 'trial',
        TimeScale.RUN: 'run',
        TimeScale.LIFE: 'life'
    }

    def __init__(self, time_step=0, pass_=0, trial=0, run=0, life=0):
        super().__init__(time_step=time_step, pass_=pass_, trial=trial, run=run, life=life)

    def _get_by_time_scale(self, time_scale):
        """
        Arguments
        ---------
            time_scale : :class:`TimeScale`

        Returns
        -------
            this Time's value of a TimeScale by the TimeScale enum, rather \
            than by attribute : int
        """
        return getattr(self, self._time_scale_attr_map[time_scale])

    def _set_by_time_scale(self, time_scale, value):
        """
        Arguments
        ---------
            time_scale : :class:`TimeScale`

        Sets this Time's value of a **time_scale** by the TimeScale enum,
        rather than by attribute
        """
        setattr(self, self._time_scale_attr_map[time_scale], value)

    def _increment_by_time_scale(self, time_scale):
        """
        Increments the value of **time_scale** in this Time by one
        """
        self._set_by_time_scale(time_scale, self._get_by_time_scale(time_scale) + 1)
        self._reset_by_time_scale(time_scale)

    def _reset_by_time_scale(self, time_scale):
        """
        Resets all the times for the time scale scope up to **time_scale**
        e.g. _reset_by_time_scale(TimeScale.TRIAL) will set the values for
        TimeScale.PASS and TimeScale.TIME_STEP to 0
        """
        for relative_time_scale in TimeScale:
            # this works because the enum is set so that higher granularities of time have lower values
            if relative_time_scale >= time_scale:
                continue

            self._set_by_time_scale(relative_time_scale, 0)


class SimpleTime(types.SimpleNamespace):
    """
    A subset class of `Time`, used to provide simple access to only
    `run <Time.run>`, `trial <Time.trial>`, and `time_step <Time.time_step>`
    """
    def __init__(self, run=0, trial=0, time_step=0):
        super().__init__(run=0, trial=0, time_step=0)

    # override __repr__ because this class is used only for cosmetic simplicity
    # based on a Time object
    def __repr__(self):
        return 'Time(run: {0}, trial: {1}, time_step: {2})'.format(self.run, self.trial, self.time_step)


class TimeHistoryTree:
    """
    A tree object that stores a history of time that has occurred at various
    :class:`TimeScale`\\ s, typically used in conjunction with a `Clock`

    Attributes
    ----------
        time_scale : :class:`TimeScale` : `TimeScale.LIFE`
            the TimeScale unit this tree/node represents

        child_time_scale : :class:`TimeScale` : `TimeScale.RUN`
            the TimeScale unit for this tree's children

        children : list[`TimeHistoryTree`]
            an ordered list of this tree's children

        max_depth : :class:`TimeScale` : `TimeScale.TRIAL`
            the finest grain TimeScale that should be created as a subtree
            Setting this value lower allows for more precise measurements
            (by default, you cannot query the number of
            `TimeScale.TIME_STEP`\\ s in a certain `TimeScale.PASS`), but
            this may use a large amount of memory in large simulations

        index : int
            the index this tree has in its parent's children list

        parent : `TimeHistoryTree` : None
            the parent node of this tree, if it exists. \
            None represents no parent (i.e. root node)

        previous_time : `Time`
            a `Time` object that represents the last time that has occurred in the tree

        current_time : `Time`
            a `Time` object that represents the current time in the tree

        total_times : dict{:class:`TimeScale`: int}
            stores the total number of units of :class:`TimeScale`\\ s that have \
            occurred over this tree's scope. Only contains entries for \
            :class:`TimeScale`\\ s of finer grain than **time_scale**

    Arguments
    ---------
        enable_current_time : bool : True
            sets this tree to maintain a `Time` object. If this tree is not
            a root (i.e. **time_scale** is `TimeScale.LIFE`)
    """
    def __init__(
        self,
        time_scale=TimeScale.LIFE,
        max_depth=TimeScale.TRIAL,
        index=0,
        parent=None,
        enable_current_time=True
    ):
        if enable_current_time:
            self.current_time = Time()
            self.previous_time = None
        self.index = index
        self.time_scale = time_scale
        self.max_depth = max_depth
        self.parent = parent

        self.child_time_scale = TimeScale.get_child(time_scale)

        if self.child_time_scale >= max_depth:
            self.children = [
                TimeHistoryTree(
                    self.child_time_scale,
                    max_depth=max_depth,
                    index=0,
                    parent=self,
                    enable_current_time=False
                )
            ]
        else:
            self.children = []

        self.total_times = {ts: 0 for ts in TimeScale if ts < self.time_scale}

    def increment_time(self, time_scale):
        """
        Increases this tree's **current_time** by one **time_scale**

        Arguments
        ---------
            time_scale : :class:`TimeScale`
                the unit of time to increment
        """
        if self.child_time_scale >= self.max_depth:
            if time_scale == self.child_time_scale:
                self.children.append(
                    TimeHistoryTree(
                        self.child_time_scale,
                        max_depth=self.max_depth,
                        index=len(self.children),
                        parent=self,
                        enable_current_time=False
                    )
                )
            else:
                self.children[-1].increment_time(time_scale)
        self.total_times[time_scale] += 1
        try:
            self.previous_time = copy.copy(self.current_time)
            self.current_time._increment_by_time_scale(time_scale)
        except AttributeError:
            # not all of these objects have time tracking
            pass

    def get_total_times_relative(
        self,
        query_time_scale,
        base_indices=None
    ):
        """
        Arguments
        ---------
            query_time_scale : :class:`TimeScale`
                the :class:`TimeScale` of units to be returned

            base_indices : dict{:class:`TimeScale`: int}
                a dictionary specifying what scope of time query_time_scale \
                is over. e.g.

                    base_indices = {TimeScale.RUN: 1, TimeScale.TRIAL: 5}

                gives the number of **query_time_scale**\\ s that have occurred \
                in the 5th `TRIAL <TimeScale.TRIAL>` of the 1st `RUN`. If an entry for a :class:`TimeScale` \
                is not specified but is coarser than **query_time_scale**, the latest \
                value for that entry will be used

        Returns
        -------
            the number of units of query_time_scale that have occurred within \
            the scope of time specified by base_indices : int
        """
        if query_time_scale >= self.time_scale:
            raise TimeScaleError(
                'query_time_scale (given: {0}) must be of finer grain than {1}.time_scale ({2})'.format(
                    query_time_scale, self, self.time_scale
                )
            )

        try:
            self.current_time
        except AttributeError:
            raise TimeScaleError(
                'get_total_times_relative should only be called on a TimeHistoryTree with enable_current_time set to True'
            )

        default_base_indices = {
            TimeScale.LIFE: 0,
            TimeScale.RUN: None,
            TimeScale.TRIAL: None,
            TimeScale.PASS: None,
        }

        # overwrite defaults with dictionary passed in argument
        if base_indices is None:
            base_indices = default_base_indices
        else:
            default_base_indices.update(base_indices)
            base_indices = default_base_indices

        base_time_scale = TimeScale.LIFE
        # base_time_scale is set as the finest grain TimeScale that is specified,
        # but more coarse than query_time_scale
        # this will be where the query to attribute times will be made
        for ts in sorted(base_indices, reverse=True):
            if base_indices[ts] is not None and ts > query_time_scale:
                base_time_scale = ts

        if base_time_scale > self.time_scale:
            raise TimeScaleError(
                'base TimeScale set by base_indices ({0}) must be at least as fine as this TimeHistoryTree\'s time_scale ({1})'.format(
                    base_time_scale,
                    self.time_scale
                )
            )

        # get the root node, which will (and should) have TimeScale.LIFE
        node = self
        while node.parent is not None:
            node = node.parent

        try:
            # for all non-specified (i.e. set to None) TimeScales coarser than base_time_scale,
            # assign them to their latest time values as default
            while node.time_scale > base_time_scale:
                if base_indices[node.child_time_scale] is None:
                    base_indices[node.child_time_scale] = len(node.children) - 1
                node = node.children[base_indices[node.child_time_scale]]

            # attempt to retrieve the correct time count given the base_indices dictionary
            node = self
            while node.time_scale != base_time_scale:
                node = node.children[base_indices[node.child_time_scale]]
            return node.total_times[query_time_scale]
        except IndexError:
            raise TimeScaleError(
                'TimeHistoryTree {0}: {1} {2} does not exist in {3} {4}'.format(
                    self,
                    node.child_time_scale,
                    base_indices[node.child_time_scale],
                    node.time_scale,
                    node.index
                )
            )



import numpy as np
from psyneulink import *


num_trials = 4
reward = 0
done = False

# These should probably be replaced by reference to ForagerEnv constants:
obs_len = 2
action_len = 2
player_coord_idx = slice(0,2)
predator_coord_idx = slice(3,5)
prey_coord_idx = slice(6,8)
player_value_idx = 2
predator_value_idx = 5
prey_value_idx = 8

player_len = prey_len = predator_len = obs_len

player = ProcessingMechanism(size=prey_len, function=GaussianDistort, name="PLAYER OBS")
prey = ProcessingMechanism(size=prey_len, function=GaussianDistort, name="PREY OBS")

# For future use:
values = TransferMechanism(size=3, name="AGENT VALUES")
reward = TransferMechanism(name="REWARD")

# Use ComparatorMechanism to compute direction of action as difference of coordinates between player and prey:
# note: unitization is done in main loop, to allow compilation of LinearCombination function) (TBI)
greedy_action_mech = ComparatorMechanism(name='MOTOR OUTPUT',sample=player,target=prey)

agent_comp = Composition(name='PREDATOR-PREY COMPOSITION')

agent_comp.add_node(player)
agent_comp.add_node(prey)
agent_comp.add_node(greedy_action_mech)

# Projections to greedy_action_mech were created by assignments of sample and target args in its constructor,
#  so just add them to the Composition).
for projection in greedy_action_mech.projections:
    agent_comp.add_projection(projection)

def main():
    for _ in range(num_trials):
        observation = np.zeros((9,))
        for i in range(4):
            run_results = agent_comp.run(inputs={
                player:[observation[player_coord_idx]],
                prey:[observation[prey_coord_idx]],
            },
            bin_execute='LLVM'
            )
            observation = np.zeros((9,))
            i+=1


if __name__ == "__main__":
    main()



import numpy as np
import pytest

import psyneulink.core.llvm as pnlvm
import psyneulink.core.components.functions.distributionfunctions as Functions

np.random.seed(0)
test_var = np.random.rand()

RAND1 = np.random.rand()
RAND2 = np.random.rand()
RAND3 = np.random.rand()
RAND4 = np.random.rand()
RAND5 = np.random.rand()

test_data = [
    (Functions.DriftDiffusionAnalytical, test_var, {}, None,
     (1.9774974807292212, 0.012242689689501842, 1.9774974807292207, 1.3147677945132479, 1.7929299891370192, 1.9774974807292207, 1.3147677945132479, 1.7929299891370192)),
    (Functions.DriftDiffusionAnalytical, test_var, {"drift_rate": RAND1, "threshold": RAND2, "starting_point": RAND3, "t0":RAND4, "noise": RAND5}, None,
     (0.4236547993389047, -2.7755575615628914e-17, 0.5173675420165031, 0.06942854144616283, 6.302631815990666, 1.4934079600147951, 0.4288991185241868, 1.7740760781361433)),
    (Functions.DriftDiffusionAnalytical, -test_var, {"drift_rate": RAND1, "threshold": RAND2, "starting_point": RAND3, "t0":RAND4, "noise": RAND5}, None,
     (0.42365479933890504, 0.0, 0.5173675420165031, 0.06942854144616283, 6.302631815990666, 1.4934079600147951, 0.4288991185241868, 1.7740760781361433)),
#    FIXME: Rounding errors result in different behaviour on different platforms
#    (Functions.DriftDiffusionAnalytical, 1e-4, {"drift_rate": 1e-5, "threshold": RAND2, "starting_point": RAND3, "t0":RAND4, "noise": RAND5}, "Rounding errors",
#     (0.5828813465336954, 0.04801236718458773, 0.532471083815943, 0.09633801362499317, 6.111833139205608, 1.5821207676710864, 0.5392724012504414, 1.8065252817609618)),
    # Two tests with different inputs to show that input is ignored.
    (Functions.NormalDist, 1e14, {"mean": RAND1, "standard_deviation": RAND2}, None, (1.0890232855122397)),
    (Functions.NormalDist, 1e-4, {"mean": RAND1, "standard_deviation": RAND2}, None, (1.0890232855122397)),
]

# use list, naming function produces ugly names
names = [
    "DriftDiffusionAnalytical-DefaultParameters",
    "DriftDiffusionAnalytical-RandomParameters",
    "DriftDiffusionAnalytical-NegInput",
#    "DriftDiffusionAnalytical-SmallDriftRate",
    "NormalDist1",
    "NormalDist2",
]

@pytest.mark.function
@pytest.mark.transfer_function
@pytest.mark.benchmark
@pytest.mark.parametrize("func, variable, params, llvm_skip, expected", test_data, ids=names)
@pytest.mark.parametrize("mode", [
    'Python',
    pytest.param('LLVM', marks=pytest.mark.llvm),
    pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda])])
def test_execute(func, variable, params, llvm_skip, expected, benchmark, mode):
    benchmark.group = "TransferFunction " + func.componentName
    if mode != 'Python' and llvm_skip:
        pytest.skip(llvm_skip)
    f = func(default_variable=variable, **params)
    if mode == 'Python':
        ex = f
    elif mode == 'LLVM':
        ex = pnlvm.execution.FuncExecution(f).execute
    elif mode == 'PTX':
        ex = pnlvm.execution.FuncExecution(f).cuda_execute
    res = ex(variable)
    assert np.allclose(res, expected)
    benchmark(f.function, variable)



import numpy as np

from psyneulink.core.components.functions.distributionfunctions import NormalDist
from psyneulink.core.components.functions.statefulfunctions.integratorfunctions import AccumulatorIntegrator
from psyneulink.core.components.mechanisms.processing.processingmechanism import ProcessingMechanism
from psyneulink.core.components.mechanisms.processing.transfermechanism import TransferMechanism
from psyneulink.core.components.projections.pathway.mappingprojection import MappingProjection
from psyneulink.core.compositions.composition import Composition
from psyneulink.core.globals.keywords import MATRIX


class TestAccumulator():

    def test_accumulator_standalone(self):
        A = AccumulatorIntegrator()
        val = A()
        assert np.allclose([[0]], val)

    def test_accumulator_initializer_len_3(self):
        A = AccumulatorIntegrator(default_variable=[[0.0], [1.0], [2.0]],
                                  initializer=[[0.0], [1.0], [2.0]],
                                  increment=1)
        assert np.allclose(A(), [[1],[2],[3]])
        assert np.allclose(A(), [[2],[3],[4]])
        assert np.allclose(A(), [[3],[4],[5]])

    # RATE ------------------------------------------------------------

    def test_accumulator_standalone_rate_float(self):
        A = AccumulatorIntegrator(rate = 0.1, increment=1)
        A()
        A()
        A()
        val = A()
        assert np.allclose([[1.111]], val)

    def test_accumulator_standalone_rate_list(self):
        A = AccumulatorIntegrator(rate = [0.1, 0.5, 0.9],
                                  increment=1)
        A()
        A()
        A()
        val = A()
        assert np.allclose([[1.111, 1.875, 3.439]], val)

    def test_accumulator_standalone_rate_ndarray(self):
        A = AccumulatorIntegrator(rate = [0.1, 0.5, 0.9],
                                  increment=1)
        A()
        A()
        A()
        val = A()
        assert np.allclose([[1.111, 1.875, 3.439]], val)

    # INCREMENT ------------------------------------------------------------

    def test_accumulator_standalone_increment_float(self):
        A = AccumulatorIntegrator(increment = 0.1)
        A()
        A()
        A()
        val = A()
        assert np.allclose([[0.4]], val)

    def test_accumulator_standalone_increment_list(self):
        A = AccumulatorIntegrator(increment = [0.1, 0.5, 0.9])
        A()
        A()
        A()
        val = A()
        assert np.allclose([[0.4, 2., 3.6]], val)

    def test_accumulator_standalone_increment_ndarray(self):
        A = AccumulatorIntegrator(increment = np.array([0.1, 0.5, 0.9]))
        A()
        A()
        A()
        val = A()
        assert np.allclose([[0.4, 2., 3.6]], val)

    # RATE X INCRMEMNT -------------------------------------------------

    def test_accumulator_standalone_rate_float_incement_float(self):
        A = AccumulatorIntegrator(
                rate=2.0,
                increment = 3.0)
        A()
        A()
        A()
        val = A()
        assert np.allclose([[45]], val)

    def test_accumulator_standalone_rate_float_incement_list(self):
        A = AccumulatorIntegrator(
                rate=2.0,
                increment = [3.5, 4.0, 5.0])
        A()
        A()
        A()
        val = A()
        assert np.allclose([[52.5, 60., 75.]], val)

    def test_accumulator_standalone_rate_list_incement_float(self):
        A = AccumulatorIntegrator(
                rate=[2.0, 2.5, -3.0],
                increment = -1.4)
        A()
        A()
        A()
        val = A()
        assert np.allclose([[-21., -35.525, 28.]], val)

    def test_accumulator_standalone_rate_list_incement_list(self):
        A = AccumulatorIntegrator(
                rate=[-.5, 2.0, -3.5],
                increment = [-1.0, .5, -2.0])
        A()
        A()
        A()
        val = A()
        assert np.allclose([[-0.625, 7.5, 66.25]], val)

    # NOISE ------------------------------------------------------------

    def test_accumulator_standalone_noise_float(self):
        A = AccumulatorIntegrator(rate = 1.0, noise=10.0)
        A()
        A()
        A()
        val = A()
        assert np.allclose([[40]], val)

    def test_accumulator_standalone_noise_list(self):
        A = AccumulatorIntegrator(rate = 1.0, noise=[10.0, 20.0, 30.0])
        A()
        A()
        A()
        val = A()
        assert np.allclose([[40., 80., 120.]], val)

    def test_accumulator_standalone_noise_ndarray(self):
        A = AccumulatorIntegrator(rate = 1.0, noise=[10.0, 20.0, 30.0])
        A()
        A()
        A()
        val = A()
        assert np.allclose([[40., 80., 120.]], val)

    def test_accumulator_standalone_noise_function(self):
        A = AccumulatorIntegrator(rate = 1.0, noise=NormalDist(standard_deviation=0.1))
        A()
        A()
        A()
        val = A()
        assert np.allclose([[-0.34591555]], val)

    def test_accumulator_standalone_noise_function_in_array(self):
        A = AccumulatorIntegrator(noise=[10, NormalDist(standard_deviation=0.1), 20])
        A()
        A()
        A()
        val = A()
        expected_val = [[40.0, 0.2480800486427607, 80.0]]
        for i in range(len(val)):
            for j in range(len(val[i])):
                assert np.allclose(expected_val[i][j], val[i][j])

    def test_accumulator_as_function_of_processing_mech(self):

        P = ProcessingMechanism(function=AccumulatorIntegrator(initializer=[0.0],
                                                               rate=.02,
                                                               increment=1))
        P.execute(1.0)
        P.execute(1.0)
        val = P.execute(1.0)
        assert np.allclose(val, [[1.0204]])

    def test_accumulator_as_function_of_matrix_param_of_mapping_projection(self):
        # Test that accumulator is function of parameter_port of mapping project,
        # and that its increment param works properly (used as modulatory param by LearningProjetion)

        T1 = TransferMechanism(size=3)
        T2 = TransferMechanism(size=3)
        M = MappingProjection(sender=T1, receiver=T2)
        C = Composition()
        C.add_linear_processing_pathway([T1, M, T2])
        C.run(inputs={T1: [1.0, 1.0, 1.0]})
        assert np.allclose(M.matrix, [[ 1.,  0.,  0.], [ 0.,  1.,  0.],[ 0.,  0.,  1.]])
        M.parameter_ports[MATRIX].function.parameters.increment.set(2, C)
        C.run(inputs={T1: [1.0, 1.0, 1.0]})
        assert np.allclose(M.matrix, [[ 3.,  2.,  2.], [ 2.,  3.,  2.], [ 2.,  2.,  3.]])
        C.run(inputs={T1: [1.0, 1.0, 1.0]})
        assert np.allclose(M.matrix, [[ 5.,  4.,  4.], [ 4.,  5.,  4.], [ 4.,  4.,  5.]])



import numpy as np
import psyneulink as pnl

import psyneulink.core.components.functions.distributionfunctions
import psyneulink.core.components.functions.statefulfunctions.integratorfunctions
import psyneulink.core.components.functions.transferfunctions
import psyneulink.core.globals.utilities

class TestModels:

    def test_DDM(self):
        myMechanism = pnl.DDM(
            function=psyneulink.core.components.functions.distributionfunctions.DriftDiffusionAnalytical(
                drift_rate=(1.0),
                threshold=(10.0),
                starting_point=0.0,
            ),
            name='My_DDM',
        )

        myMechanism_2 = pnl.DDM(
            function=psyneulink.core.components.functions.distributionfunctions.DriftDiffusionAnalytical(
                drift_rate=2.0,
                threshold=20.0),
            name='My_DDM_2'
        )

        myMechanism_3 = pnl.DDM(
            function=psyneulink.core.components.functions.distributionfunctions.DriftDiffusionAnalytical(
                drift_rate=3.0,
                threshold=30.0
            ),
            name='My_DDM_3',
        )

        z = pnl.Composition()
        z.add_linear_processing_pathway([myMechanism,
                                        pnl.MappingProjection(matrix=pnl.IDENTITY_MATRIX),
                                        myMechanism_2,
                                         pnl.MappingProjection(matrix=pnl.FULL_CONNECTIVITY_MATRIX),
                                        myMechanism_3])

        result = z.run(inputs={myMechanism: [[40]]})[0][0]

        expected_output = [
            (myMechanism.input_ports[0].parameters.value.get(z), np.array([40.])),
            (myMechanism.output_ports[0].parameters.value.get(z), np.array([10.])),
            (myMechanism_2.input_ports[0].parameters.value.get(z), np.array([10.])),
            (myMechanism_2.output_ports[0].parameters.value.get(z), np.array([20.])),
            (myMechanism_3.input_ports[0].parameters.value.get(z), np.array([20.])),
            (myMechanism_3.output_ports[0].parameters.value.get(z), np.array([30.])),
            (result, np.array([30.])),
        ]

        for i in range(len(expected_output)):
            val, expected = expected_output[i]
            # setting absolute tolerance to be in accordance with reference_output precision
            # if you do not specify, assert_allcose will use a relative tolerance of 1e-07,
            # which WILL FAIL unless you gather higher precision values to use as reference
            np.testing.assert_allclose(val, expected, atol=1e-08, err_msg='Failed on expected_output[{0}]'.format(i))

    # FIX: 5/8/20 [JDC] Needs assertions
    def test_bustamante_Stroop_model(self):
        #  INPUT UNITS

        #  colors: ('red', 'green'), words: ('RED','GREEN')
        colors_input_layer = pnl.TransferMechanism(size=2,
                                                   function=psyneulink.core.components.functions.transferfunctions.Linear,
                                                   name='COLORS_INPUT')

        words_input_layer = pnl.TransferMechanism(size=2,
                                                  function=psyneulink.core.components.functions.transferfunctions.Linear,
                                                  name='WORDS_INPUT')

        #   Task layer, tasks: ('name the color', 'read the word')
        task_layer = pnl.TransferMechanism(size=2,
                                           function=psyneulink.core.components.functions.transferfunctions.Linear,
                                           name='TASK')

        #   HIDDEN LAYER UNITS

        #   colors_hidden: ('red','green')
        #   Logistic activation function, Gain = 1.0, Bias = -4.0 (in PNL bias is subtracted so enter +4.0 to get negative bias)
        #   randomly distributed noise to the net input
        #   time averaging = integration_rate = 0.1
        unit_noise = 0.005
        colors_hidden_layer = pnl.TransferMechanism(size=2,
                                                    function=psyneulink.core.components.functions.transferfunctions
                                                    .Logistic(gain=1.0, x_0=4.0),
                                                    # should be able to get same result with offset = -4.0
                                                    integrator_mode=True,
                                                    noise=psyneulink.core.components.functions.distributionfunctions
                                                    .NormalDist(mean=0, standard_deviation=unit_noise).function,
                                                    integration_rate=0.1,
                                                    name='COLORS HIDDEN')
        #    words_hidden: ('RED','GREEN')
        words_hidden_layer = pnl.TransferMechanism(size=2,
                                                   function=pnl.Logistic(gain=1.0, x_0=4.0),
                                                   integrator_mode=True,
                                                   noise=pnl.NormalDist(mean=0,
                                                                        standard_deviation=unit_noise).function,
                                                   integration_rate=0.1,
                                                   name='WORDS HIDDEN')

        #    OUTPUT UNITS

        #   Response layer, provide input to accumulator, responses: ('red', 'green')
        #   time averaging = tau = 0.1
        #   randomly distributed noise to the net input
        response_layer = pnl.TransferMechanism(size=2,
                                               function=psyneulink.core.components.functions.transferfunctions.Logistic,
                                               name='RESPONSE',
                                               integrator_mode=True,
                                               noise=psyneulink.core.components.functions.distributionfunctions.NormalDist(mean=0, standard_deviation=unit_noise).function,
                                               integration_rate=0.1)
        #   Respond red accumulator
        #   alpha = rate of evidence accumlation = 0.1
        #   sigma = noise = 0.1
        #   noise will be: squareroot(time_step_size * noise) * a random sample from a normal distribution
        accumulator_noise = 0.1
        respond_red_accumulator = pnl.IntegratorMechanism(function=pnl.SimpleIntegrator(noise=pnl.NormalDist(mean=0,
                                                                                                             standard_deviation=accumulator_noise).function,
                                                                                        rate=0.1),
                                                          name='respond_red_accumulator')
        #   Respond green accumulator
        respond_green_accumulator = pnl.IntegratorMechanism(function=pnl.SimpleIntegrator(noise=pnl.NormalDist(mean=0,
                                                                                                               standard_deviation=accumulator_noise).function,
                                                                                          rate=0.1),
                                                            name='respond_green_accumulator')

        #   LOGGING
        colors_hidden_layer.set_log_conditions('value')
        words_hidden_layer.set_log_conditions('value')
        response_layer.set_log_conditions('value')
        respond_red_accumulator.set_log_conditions('value')
        respond_green_accumulator.set_log_conditions('value')

        #   SET UP CONNECTIONS

        #   rows correspond to sender
        #   columns correspond to: weighting of the contribution that a given sender makes to the receiver

        #   INPUT TO HIDDEN
        # row 0: input_'red' to hidden_'red', hidden_'green'
        # row 1: input_'green' to hidden_'red', hidden_'green'
        color_weights = pnl.MappingProjection(matrix=np.atleast_2d([[2.2, -2.2],
                                                                [-2.2, 2.2]]),
                                              name='COLOR_WEIGHTS')
        # row 0: input_'RED' to hidden_'RED', hidden_'GREEN'
        # row 1: input_'GREEN' to hidden_'RED', hidden_'GREEN'
        word_weights = pnl.MappingProjection(matrix=np.atleast_2d([[2.6, -2.6],
                                                               [-2.6, 2.6]]),
                                             name='WORD_WEIGHTS')

        #   HIDDEN TO RESPONSE
        # row 0: hidden_'red' to response_'red', response_'green'
        # row 1: hidden_'green' to response_'red', response_'green'
        color_response_weights = pnl.MappingProjection(matrix=np.atleast_2d([[1.3, -1.3],
                                                                         [-1.3, 1.3]]),
                                                       name='COLOR_RESPONSE_WEIGHTS')
        # row 0: hidden_'RED' to response_'red', response_'green'
        # row 1: hidden_'GREEN' to response_'red', response_'green'
        word_response_weights = pnl.MappingProjection(matrix=np.atleast_2d([[2.5, -2.5],
                                                                        [-2.5, 2.5]]),
                                                      name='WORD_RESPONSE_WEIGHTS')

        #   TASK TO HIDDEN LAYER
        #   row 0: task_CN to hidden_'red', hidden_'green'
        #   row 1: task_WR to hidden_'red', hidden_'green'
        task_CN_weights = pnl.MappingProjection(matrix=np.atleast_2d([[4.0, 4.0],
                                                                  [0, 0]]),
                                                name='TASK_CN_WEIGHTS')

        #   row 0: task_CN to hidden_'RED', hidden_'GREEN'
        #   row 1: task_WR to hidden_'RED', hidden_'GREEN'
        task_WR_weights = pnl.MappingProjection(matrix=np.atleast_2d([[0, 0],
                                                                  [4.0, 4.0]]),
                                                name='TASK_WR_WEIGHTS')

        #   RESPONSE UNITS TO ACCUMULATORS
        #   row 0: response_'red' to respond_red_accumulator
        #   row 1: response_'green' to respond_red_accumulator
        respond_red_differencing_weights = pnl.MappingProjection(matrix=np.atleast_2d([[1.0], [-1.0]]),
                                                                 name='RESPOND_RED_WEIGHTS')

        #   row 0: response_'red' to respond_green_accumulator
        #   row 1: response_'green' to respond_green_accumulator
        respond_green_differencing_weights = pnl.MappingProjection(matrix=np.atleast_2d([[-1.0], [1.0]]),
                                                                   name='RESPOND_GREEN_WEIGHTS')

        # CREATE COMPOSITION FROM PATHWAYS
        my_Stroop = pnl.Composition(pathways=[
            {'WORD_PATHWAY':[words_input_layer,
                             word_weights,
                             words_hidden_layer,
                             word_response_weights,
                             response_layer]},
            {'COLO_PATHWAY': [colors_input_layer,
                              color_weights,
                              colors_hidden_layer,
                              color_response_weights,
                              response_layer]},
            {'TASK_CN_PATHWAY':[task_layer,
                                task_CN_weights,
                                colors_hidden_layer]},
            {'TASK_WR_PATHWAY':[task_layer,
                                task_WR_weights,
                                words_hidden_layer]},
            {'RESPOND_RED_PATHWAY':[response_layer,
                                    respond_red_differencing_weights,
                                    respond_red_accumulator]},
            {'RESPOND_GREEN_PATHWAY': [response_layer,
                                       respond_green_differencing_weights,
                                       respond_green_accumulator]},
        ])

        # my_Stroop.show()
        # my_Stroop.show_graph(show_dimensions=pnl.ALL)

        # Function to create test trials
        # a RED word input is [1,0] to words_input_layer and GREEN word is [0,1]
        # a red color input is [1,0] to colors_input_layer and green color is [0,1]
        # a color-naming trial is [1,0] to task_layer and a word-reading trial is [0,1]

        def trial_dict(red_color, green_color, red_word, green_word, CN, WR):

            trialdict = {
                colors_input_layer: [red_color, green_color],
                words_input_layer: [red_word, green_word],
                task_layer: [CN, WR]
            }
            return trialdict

        #   CREATE THRESHOLD FUNCTION
        # first value of DDM's value is DECISION_VARIABLE
        # context is always passed to Condition functions and is the context
        # in which the function gets called - below, during system execution
        def pass_threshold(mech1, mech2, thresh, context=None):
            results1 = mech1.output_ports[0].parameters.value.get(context)
            results2 = mech2.output_ports[0].parameters.value.get(context)
            for val in results1:
                if val >= thresh:
                    return True
            for val in results2:
                if val >= thresh:
                    return True
            return False

        accumulator_threshold = 1.0

        mechanisms_to_update = [colors_hidden_layer, words_hidden_layer, response_layer]

        def switch_integrator_mode(mechanisms, mode):
            for mechanism in mechanisms:
                mechanism.integrator_mode = mode

        def switch_noise(mechanisms, noise):
            for mechanism in mechanisms:
                mechanism.noise = noise

        def switch_to_initialization_trial(mechanisms):
            # Turn off accumulation
            switch_integrator_mode(mechanisms, False)
            # Turn off noise
            switch_noise(mechanisms, 0)
            # Execute once per trial
            my_Stroop.termination_processing = {pnl.TimeScale.TRIAL: pnl.AllHaveRun()}

        def switch_to_processing_trial(mechanisms):
            # Turn on accumulation
            switch_integrator_mode(mechanisms, True)
            # Turn on noise
            switch_noise(mechanisms, psyneulink.core.components.functions.distributionfunctions.NormalDist(mean=0, standard_deviation=unit_noise).function)
            # Execute until one of the accumulators crosses the threshold
            my_Stroop.termination_processing = {
                pnl.TimeScale.TRIAL: pnl.While(
                    pass_threshold,
                    respond_red_accumulator,
                    respond_green_accumulator,
                    accumulator_threshold
                )
            }

        def switch_trial_type():
            # Next trial will be a processing trial
            if isinstance(my_Stroop.termination_processing[pnl.TimeScale.TRIAL], pnl.AllHaveRun):
                switch_to_processing_trial(mechanisms_to_update)
            # Next trial will be an initialization trial
            else:
                switch_to_initialization_trial(mechanisms_to_update)

        CN_trial_initialize_input = trial_dict(0, 0, 0, 0, 1, 0)

        WR_trial_initialize_input = trial_dict(0, 0, 0, 0, 0, 1)

        # Start with an initialization trial
        switch_to_initialization_trial(mechanisms_to_update)

        my_Stroop.run(inputs=trial_dict(0, 1, 1, 0, 1, 0),
                      # termination_processing=change_termination_processing,
                      num_trials=4,
                      call_after_trial=switch_trial_type)

        # {colors_input_layer: [[0, 0], [1, 0]],
        #                       words_input_layer: [[0, 0], [1, 0]],
        #                       task_layer: [[0, 1], [0, 1]]}


    # This script implements Figure 1 of Botvinick, M. M., Braver, T. S., Barch, D. M., Carter, C. S., & Cohen, J. D. (2001).
    # Conflict monitoring and cognitive control. Psychological Review, 108, 624–652.
    # http://dx.doi.org/10.1037/0033-295X.108.3.624

    # Figure 1 plots the ENERGY computed by a conflict mechanism. It is highest for incongruent trials,
    # and similar for congruent and neutral trials.
    # Noise is turned of and for each condition we ran one trial only. A response threshold was not defined. Responses were
    # made at the marked * signs in the figure.
    # Note that this script implements a slightly different Figure than in the original Figure in the paper.
    # However, this implementation is identical with a plot we created with an old MATLAB code which was used for the
    # conflict monitoring simulations.


    # def test_botvinick_model(self):
    #
    #     colors_input_layer = pnl.TransferMechanism(size=3,
    #                                                function=pnl.Linear,
    #                                                name='COLORS_INPUT')
    #
    #     words_input_layer = pnl.TransferMechanism(size=3,
    #                                               function=pnl.Linear,
    #                                               name='WORDS_INPUT')
    #
    #     task_input_layer = pnl.TransferMechanism(size=2,
    #                                              function=pnl.Linear,
    #                                              name='TASK_INPUT')
    #
    #     task_layer = pnl.RecurrentTransferMechanism(size=2,
    #                                                 function=pnl.Logistic(),
    #                                                 hetero=-2,
    #                                                 integrator_mode=True,
    #                                                 integration_rate=0.01,
    #                                                 name='TASK_LAYER')
    #
    #     colors_hidden_layer = pnl.RecurrentTransferMechanism(size=3,
    #                                                          function=pnl.Logistic(bias=4.0),
    #                                                          # bias 4.0 is -4.0 in the paper see Docs for description
    #                                                          integrator_mode=True,
    #                                                          hetero=-2,
    #                                                          integration_rate=0.01,  # cohen-huston text says 0.01
    #                                                          name='COLORS_HIDDEN')
    #
    #     words_hidden_layer = pnl.RecurrentTransferMechanism(size=3,
    #                                                         function=pnl.Logistic(bias=4.0),
    #                                                         integrator_mode=True,
    #                                                         hetero=-2,
    #                                                         integration_rate=0.01,
    #                                                         name='WORDS_HIDDEN')
    #
    #     #   Response layer, responses: ('red', 'green')
    #     response_layer = pnl.RecurrentTransferMechanism(size=2,
    #                                                     function=pnl.Logistic(),
    #                                                     hetero=-2.0,
    #                                                     integrator_mode=True,
    #                                                     integration_rate=0.01,
    #                                                     output_ports=[pnl.RESULT,
    #                                                                    {pnl.NAME: 'DECISION_ENERGY',
    #                                                                     pnl.VARIABLE: (pnl.OWNER_VALUE, 0),
    #                                                                     pnl.FUNCTION: pnl.Stability(
    #                                                                         default_variable=np.array([0.0, 0.0]),
    #                                                                         metric=pnl.ENERGY,
    #                                                                         matrix=np.array([[0.0, -4.0],
    #                                                                                          [-4.0, 0.0]]))}],
    #                                                     name='RESPONSE', )
    #
    #     response_layer.set_log_conditions('DECISION_ENERGY')
    #
    #     color_input_weights = pnl.MappingProjection(matrix=np.array([[1.0, 0.0, 0.0],
    #                                                                  [0.0, 1.0, 0.0],
    #                                                                  [0.0, 0.0, 1.0]]))
    #
    #     word_input_weights = pnl.MappingProjection(matrix=np.array([[1.0, 0.0, 0.0],
    #                                                                 [0.0, 1.0, 0.0],
    #                                                                 [0.0, 0.0, 1.0]]))
    #
    #     task_input_weights = pnl.MappingProjection(matrix=np.array([[1.0, 0.0],
    #                                                                 [0.0, 1.0]]))
    #
    #     color_task_weights = pnl.MappingProjection(matrix=np.array([[4.0, 0.0],
    #                                                                 [4.0, 0.0],
    #                                                                 [4.0, 0.0]]))
    #
    #     task_color_weights = pnl.MappingProjection(matrix=np.array([[4.0, 4.0, 4.0],
    #                                                                 [0.0, 0.0, 0.0]]))
    #
    #     response_color_weights = pnl.MappingProjection(matrix=np.array([[1.5, 0.0, 0.0],
    #                                                                     [0.0, 1.5, 0.0]]))
    #
    #     response_word_weights = pnl.MappingProjection(matrix=np.array([[2.5, 0.0, 0.0],
    #                                                                    [0.0, 2.5, 0.0]]))
    #
    #     color_response_weights = pnl.MappingProjection(matrix=np.array([[1.5, 0.0],
    #                                                                     [0.0, 1.5],
    #                                                                     [0.0, 0.0]]))
    #
    #     word_response_weights = pnl.MappingProjection(matrix=np.array([[2.5, 0.0],
    #                                                                    [0.0, 2.5],
    #                                                                    [0.0, 0.0]]))
    #
    #     word_task_weights = pnl.MappingProjection(matrix=np.array([[0.0, 4.0],
    #                                                                [0.0, 4.0],
    #                                                                [0.0, 4.0]]))
    #
    #     task_word_weights = pnl.MappingProjection(matrix=np.array([[0.0, 0.0, 0.0],
    #                                                                [4.0, 4.0, 4.0]]))
    #
    #     # CREATE Composition
    #     comp = pnl.Composition()
    #
    #     comp.add_linear_processing_pathway([colors_input_layer,
    #                                         color_input_weights,
    #                                         colors_hidden_layer,
    #                                         color_response_weights,
    #                                         response_layer])
    #     comp.add_projection(response_color_weights, response_layer, colors_hidden_layer)
    #
    #     comp.add_linear_processing_pathway([words_input_layer,
    #                                         word_input_weights,
    #                                         words_hidden_layer,
    #                                         word_response_weights,
    #                                         response_layer])
    #     comp.add_projection(response_word_weights, response_layer, words_hidden_layer)
    #
    #     comp.add_projection(task_input_weights, task_input_layer, task_layer)
    #
    #     comp.add_projection(task_color_weights, task_layer, colors_hidden_layer)
    #     comp.add_projection(color_task_weights, colors_hidden_layer, task_layer)
    #
    #     comp.add_projection(task_word_weights, task_layer, words_hidden_layer)
    #     comp.add_projection(word_task_weights, words_hidden_layer, task_layer)
    #
    #     def trial_dict(red_color, green_color, neutral_color, red_word, green_word, neutral_word, CN, WR):
    #         trialdict = {
    #             colors_input_layer: [red_color, green_color, neutral_color],
    #             words_input_layer: [red_word, green_word, neutral_word],
    #             task_input_layer: [CN, WR]
    #         }
    #         return trialdict
    #
    #     # Define initialization trials separately
    #     CN_trial_initialize_input = trial_dict(0, 0, 0, 0, 0, 0, 1,
    #                                            0)  # red_color, green color, red_word, green word, CN, WR
    #     CN_incongruent_trial_input = trial_dict(1, 0, 0, 0, 1, 0, 1,
    #                                             0)  # red_color, green color, red_word, green word, CN, WR
    #     CN_congruent_trial_input = trial_dict(1, 0, 0, 1, 0, 0, 1,
    #                                           0)  # red_color, green color, red_word, green word, CN, WR
    #     CN_control_trial_input = trial_dict(1, 0, 0, 0, 0, 1, 1,
    #                                         0)  # red_color, green color, red_word, green word, CN, WR
    #
    #     Stimulus = [[CN_trial_initialize_input, CN_congruent_trial_input],
    #                 [CN_trial_initialize_input, CN_incongruent_trial_input],
    #                 [CN_trial_initialize_input, CN_control_trial_input]]
    #
    #     # should be 500 and 1000
    #     ntrials0 = 50
    #     ntrials = 100
    #
    #     def run():
    #         results = []
    #         for stim in Stimulus:
    #             # RUN the SYSTEM to initialize ----------------------------------------------------------------------------------------
    #             res = comp.run(inputs=stim[0], num_trials=ntrials0)
    #             results.append(res)
    #             res = comp.run(inputs=stim[1], num_trials=ntrials)
    #             results.append(res)
    #             # reset after condition was run
    #             colors_hidden_layer.reset([[0, 0, 0]])
    #             words_hidden_layer.reset([[0, 0, 0]])
    #             response_layer.reset([[0, 0]])
    #             task_layer.reset([[0, 0]])
    #             comp.reset()
    #
    #         return results
    #
    #     res = run()
    #     assert np.allclose(res[0], [0.04946301, 0.04946301, 0.03812533])
    #     assert np.allclose(res[1], [0.20351701, 0.11078586, 0.04995664])
    #     assert np.allclose(res[2], [0.04946301, 0.04946301, 0.03812533])
    #     assert np.allclose(res[3], [0.11168014, 0.20204928, 0.04996308])
    #     assert np.allclose(res[4], [0.05330691, 0.05330691, 0.03453411])
    #     assert np.allclose(res[5], [0.11327619, 0.11238362, 0.09399782])
    #
    #     if mode == 'LLVM':
    #         return
    #     r2 = response_layer.log.nparray_dictionary('DECISION_ENERGY')  # get logged DECISION_ENERGY dictionary
    #     energy = r2['DECISION_ENERGY']  # save logged DECISION_ENERGY
    #
    #     assert np.allclose(energy[:450],
    #                        [0.9907482, 0.98169891, 0.97284822, 0.96419228, 0.95572727, 0.94744946,
    #                         0.93935517, 0.93144078, 0.92370273, 0.91613752, 0.90874171, 0.90151191,
    #                         0.89444481, 0.88753715, 0.88078573, 0.8741874, 0.8677391, 0.86143779,
    #                         0.85528051, 0.84926435, 0.84338646, 0.83764405, 0.83203438, 0.82655476,
    #                         0.82120257, 0.81597521, 0.81087018, 0.805885, 0.80101724, 0.79626453,
    #                         0.79162455, 0.78709503, 0.78267373, 0.77835847, 0.77414712, 0.77003759,
    #                         0.76602783, 0.76211584, 0.75829965, 0.75457736, 0.75094707, 0.74740696,
    #                         0.74395521, 0.74059008, 0.73730983, 0.73411278, 0.73099728, 0.72796172,
    #                         0.7250045, 0.72212409, 0.71932195, 0.71660193, 0.7139626, 0.71140257,
    #                         0.70892047, 0.70651498, 0.70418478, 0.70192859, 0.69974515, 0.69763324,
    #                         0.69559166, 0.69361922, 0.69171476, 0.68987717, 0.68810533, 0.68639816,
    #                         0.68475458, 0.68317357, 0.6816541, 0.68019517, 0.67879579, 0.67745502,
    #                         0.6761719, 0.67494551, 0.67377494, 0.67265932, 0.67159776, 0.67058942,
    #                         0.66963346, 0.66872906, 0.66787541, 0.66707173, 0.66631723, 0.66561116,
    #                         0.66495278, 0.66434134, 0.66377614, 0.66325648, 0.66278164, 0.66235097,
    #                         0.66196379, 0.66161945, 0.66131731, 0.66105673, 0.66083709, 0.6606578,
    #                         0.66051824, 0.66041784, 0.66035601, 0.66033219, 0.66034583, 0.66039637,
    #                         0.66048328, 0.66060603, 0.6607641, 0.66095698, 0.66118416, 0.66144515,
    #                         0.66173948, 0.66206664, 0.66242619, 0.66281766, 0.66324058, 0.66369451,
    #                         0.66417902, 0.66469367, 0.66523802, 0.66581167, 0.66641419, 0.66704517,
    #                         0.66770422, 0.66839095, 0.66910495, 0.66984584, 0.67061326, 0.67140681,
    #                         0.67222614, 0.67307088, 0.67394067, 0.67483517, 0.67575402, 0.67669687,
    #                         0.67766339, 0.67865325, 0.67966612, 0.68070166, 0.68175955, 0.68283949,
    #                         0.68394114, 0.68506421, 0.68620839, 0.68737336, 0.68855885, 0.68976453,
    #                         0.69099013, 0.69223536, 0.69349992, 0.69478353, 0.69608592, 0.6974068,
    #                         0.9907482, 0.98169891, 0.97284822, 0.96419228, 0.95572727, 0.94744946,
    #                         0.93935517, 0.93144078, 0.92370273, 0.91613752, 0.90874171, 0.90151191,
    #                         0.89444481, 0.88753715, 0.88078573, 0.8741874, 0.8677391, 0.86143779,
    #                         0.85528051, 0.84926435, 0.84338646, 0.83764405, 0.83203438, 0.82655476,
    #                         0.82120257, 0.81597521, 0.81087018, 0.805885, 0.80101724, 0.79626453,
    #                         0.79162455, 0.78709503, 0.78267373, 0.77835847, 0.77414712, 0.77003759,
    #                         0.76602783, 0.76211584, 0.75829965, 0.75457736, 0.75094707, 0.74740696,
    #                         0.74395521, 0.74059008, 0.73730983, 0.73411278, 0.73099728, 0.72796172,
    #                         0.7250045, 0.72212409, 0.71932195, 0.71660193, 0.7139626, 0.71140257,
    #                         0.70892048, 0.70651499, 0.70418479, 0.70192861, 0.69974518, 0.69763329,
    #                         0.69559172, 0.69361931, 0.69171489, 0.68987734, 0.68810556, 0.68639845,
    #                         0.68475496, 0.68317405, 0.6816547, 0.68019591, 0.6787967, 0.67745612,
    #                         0.67617322, 0.67494709, 0.67377682, 0.67266154, 0.67160036, 0.67059245,
    #                         0.66963697, 0.6687331, 0.66788005, 0.66707703, 0.66632327, 0.66561801,
    #                         0.66496052, 0.66435007, 0.66378595, 0.66326745, 0.6627939, 0.66236463,
    #                         0.66197896, 0.66163626, 0.66133589, 0.66107724, 0.66085967, 0.66068261,
    #                         0.66054545, 0.66044762, 0.66038855, 0.66036769, 0.66038449, 0.66043841,
    #                         0.66052892, 0.66065551, 0.66081767, 0.6610149, 0.6612467, 0.6615126,
    #                         0.66181212, 0.66214479, 0.66251017, 0.6629078, 0.66333724, 0.66379805,
    #                         0.66428981, 0.66481211, 0.66536452, 0.66594665, 0.66655809, 0.66719846,
    #                         0.66786737, 0.66856444, 0.66928929, 0.67004157, 0.67082092, 0.67162697,
    #                         0.67245938, 0.67331781, 0.67420192, 0.67511137, 0.67604585, 0.67700502,
    #                         0.67798857, 0.67899619, 0.68002758, 0.68108242, 0.68216042, 0.6832613,
    #                         0.68438475, 0.68553049, 0.68669824, 0.68788774, 0.68909869, 0.69033084,
    #                         0.69158392, 0.69285767, 0.69415183, 0.69546615, 0.69680037, 0.69815425,
    #                         0.9907482, 0.98169891, 0.97284822, 0.96419228, 0.95572727, 0.94744946,
    #                         0.93935517, 0.93144078, 0.92370273, 0.91613752, 0.90874171, 0.90151191,
    #                         0.89444481, 0.88753715, 0.88078573, 0.8741874, 0.8677391, 0.86143779,
    #                         0.85528051, 0.84926435, 0.84338646, 0.83764405, 0.83203438, 0.82655476,
    #                         0.82120257, 0.81597521, 0.81087018, 0.805885, 0.80101724, 0.79626453,
    #                         0.79162455, 0.78709503, 0.78267373, 0.77835847, 0.77414712, 0.77003759,
    #                         0.76602783, 0.76211584, 0.75829965, 0.75457736, 0.75094707, 0.74740696,
    #                         0.74395521, 0.74059008, 0.73730983, 0.73411278, 0.73099728, 0.72796172,
    #                         0.7250045, 0.72212409, 0.71932195, 0.7165966, 0.71394661, 0.71137057,
    #                         0.70886708, 0.70643479, 0.70407238, 0.70177856, 0.69955206, 0.69739162,
    #                         0.69529605, 0.69326414, 0.69129474, 0.68938671, 0.68753892, 0.6857503,
    #                         0.68401976, 0.68234626, 0.68072878, 0.67916632, 0.67765789, 0.67620253,
    #                         0.67479929, 0.67344727, 0.67214554, 0.67089324, 0.66968949, 0.66853344,
    #                         0.66742427, 0.66636116, 0.66534331, 0.66436994, 0.66344028, 0.66255359,
    #                         0.66170913, 0.66090618, 0.66014404, 0.65942201, 0.65873941, 0.65809559,
    #                         0.65748989, 0.65692167, 0.6563903, 0.65589518, 0.6554357, 0.65501128,
    #                         0.65462132, 0.65426528, 0.65394258, 0.65365269, 0.65339506, 0.65316919,
    #                         0.65297454, 0.65281061, 0.65267692, 0.65257297, 0.65249828, 0.65245239,
    #                         0.65243484, 0.65244517, 0.65248294, 0.65254773, 0.65263909, 0.65275661,
    #                         0.65289989, 0.6530685, 0.65326207, 0.65348019, 0.65372249, 0.65398859,
    #                         0.65427811, 0.6545907, 0.65492599, 0.65528364, 0.6556633, 0.65606463,
    #                         0.65648729, 0.65693097, 0.65739533, 0.65788006, 0.65838485, 0.65890938,
    #                         0.65945336, 0.6600165, 0.66059849, 0.66119904, 0.66181789, 0.66245474,
    #                         0.66310932, 0.66378136, 0.6644706, 0.66517677, 0.66589961, 0.66663887,
    #                         0.66739429, 0.66816564, 0.66895265, 0.66975511, 0.67057276, 0.67140537],
    #                        atol=1e-02)
